<?php
require_once 'AlertSQL.php';
class SoftwareUpdateAlert extends AlertSQL {


    function __construct()
    {
        parent::__construct();
    }

    /**
     * Return SQL string with condition for software update
     *
     *
     * @param <CF_RULE> $rule
     * @return string
     */
    private function _getSoftwareUpdateConditions($rule)
    {
        $tmp = array();

        $condition = $rule->softwareUpdateConditions;

        $tmp[] = " PatchReportType='AVAILABLE' ";


        if (!empty($condition['patchname'])) {
            if (isset($condition['condition']) && $condition['condition'] == 'is') {
                $tmp[] = "patchname ILIKE '" . $condition['patchname'] . "'";
            } else {
                $tmp[] = "patchname ILIKE '%" . $condition['patchname'] . "%'";
            }
        }

        if (!empty($condition['patcharchitecture']))
        {
            $tmp[] = "patcharchitecture ILIKE '" . $condition['patcharchitecture'] . "'";
        }

        $WHERE_CONDITION = '';
        if (!empty($tmp))
        {
            $WHERE_CONDITION = 'WHERE ' . implode(" AND ", $tmp);
        }

        return $WHERE_CONDITION;
    }


    /**
     * Return status info for software updates alerts.
     *
     * @param <CF_ALERT> $alert
     * @param <CF_RULE> $rule
     * @param <array> $includes
     * @param <array> $excludes
     * @return <array>
     */

    function getSoftwareUpdateStatus($alert, $rule, $includes, $excludes)
    {
        $failHosts =  $this->utils_getCountFailHosts($alert, $rule, $includes, $excludes);

        $totalHosts = $this->ci->host_model->getHostCountByContext($alert->username, $includes, $excludes);

        $alertStatus['status']     = $failHosts > 0 ? 'fail' : 'success';
        $alertStatus['failHosts']  = $failHosts;
        $alertStatus['totalHosts'] = $totalHosts;

        return $alertStatus;
    }

    /**
     * Return SQL string for hostkeys
     *
     * @param <CF_RULE> $rule
     * @return type
     */
    function getStatSQLStringForSoftwareUpdate($rule) {
        $WHERE_CONDITION = $this->_getSoftwareUpdateConditions($rule);
        $SQL = "SELECT DISTINCT hostkey FROM SoftwareUpdates $WHERE_CONDITION";
        return $SQL;
    }


    /**
     * Return count of fail hosts
     *
     * @param <CF_ALERT> $alert
     * @param <CF_RULE> $rule
     * @return <int>
     */

    // NOTE: for future refactoring.
    // getCountFailHosts  and getSoftwareUpdateStatus use different includes, so in the getSoftwareUpdateStatus we can not use utils_getIncludesFromAlert function to get includes list,
    // because in this case includes is a combination of all includes from group of alerts.

    function getCountFailHosts($alert, $rule)
    {
        $includes  = $this->utils_getIncludesFromAlert($alert);
        $excludes  = $this->utils_getExcludesFromAlert($alert);
        $failHosts =  $this->utils_getCountFailHosts($alert, $rule, $includes, $excludes);

        return $failHosts;
    }

    // see note for getCountFailHosts function about WHY includes is send as param and not taken from $alert
    private function utils_getCountFailHosts($alert, $rule, $includes=array(), $excludes = [])
    {
        $WHERE_CONDITION = $this->_getSoftwareUpdateConditions($rule);

        $SQL = "SELECT COUNT(*) AS failHost
                FROM (
                    SELECT DISTINCT s_up.hostkey
                    FROM softwareUpdates s_up
                    $WHERE_CONDITION
                ) AS c_query";

        try
        {
            $result = $this->ci->advancedreports_model->runQuery($alert->username, $SQL, '', '', 0, 0, $includes, $excludes);
        }
        catch (Exception $e)
        {
            log_message('error', $e->getMessage() . " " . $e->getFile() . " line:" . $e->getLine());
            throw $e;
        }



        if (isset($result['rows']) && isset($result['rows'][0])) {
            $failHosts  = $result['rows'][0][0];
        }
        else
        {
            // result if nothing found - Assume success for now
            $failHosts  = 0;
        }

        return $failHosts;
    }

    /**
     * Return list of hosts for alerts. Note, we MUST have hostkey here to set proper context
     *
     * @param <CF_ALERT> $alert
     * @param <CF_RULE> $rule
     * @param <int> $limit
     * @return <array>
     */
    function getFailHostsList_ForSoftwareUpdateAlerts($alert, $rule, $limit = 5)
    {
        $WHERE_CONDITION = $this->_getSoftwareUpdateConditions($rule);

        $SQL = 'SELECT h.hostkey, h.hostname, count(s.PatchName) AS "c"
                FROM hosts h
                INNER JOIN SoftwareUpdates s ON s.hostkey = h.hostkey ' .
                $WHERE_CONDITION . '
                GROUP BY h.hostkey, h.hostname ORDER BY c DESC';

        $SQL .= ' LIMIT '.(int)$limit;

        $includes = $this->utils_getIncludesFromAlert($alert);
        $excludes = $this->utils_getExcludesFromAlert($alert);

        $result = $this->ci->advancedreports_model->runQuery($alert->username, $SQL, '', '', 0, '', $includes, $excludes);
        return $result['rows'];
    }



   /**
     * Return a link to the report without adding include/exclude or hostkey.
     *
     *
     * @param <CF_RULE> $rule
     * @param string $hostkey
     * @return type
     */
    function getLinkToSoftwareUpdatesReport($rule, $hostkey)
    {
        $WHERE_CONDITION = $this->_getSoftwareUpdateConditions($rule);

        if (!empty($hostkey))
        {
            if (empty($WHERE_CONDITION))
            {
                $WHERE_CONDITION = 'WHERE s.Hostkey = \'' . $hostkey .'\'';
            }
            else
            {
                $WHERE_CONDITION .= ' AND s.Hostkey = \'' . $hostkey . '\'';
            }
        }

        // create link str to use it in report
        $link =     'SELECT
                        h.hostname AS "Host name",
                        s_up.patchname AS "Package name",
                        array_agg(s.softwareversion) AS "Versions installed",
                        s_up.patchversion AS "Version available",
                        s_up.changetimestamp AS "Discovered at" 
                    FROM
                        hosts h
                    INNER
                        JOIN softwareupdates s_up ON s_up.hostkey = h.hostkey
                    INNER
                        JOIN software s ON s.hostkey = s_up.hostkey
                        AND s.softwarename = s_up.patchname
                        AND s.softwarearchitecture = s_up.patcharchitecture
                    ' . $WHERE_CONDITION . '
                    GROUP BY 
                        h.hostkey,
                        h.hostname,
                        s_up.patchname,
                        s_up.patchversion,
                        s_up.changetimestamp';

        return rawurlencode(base64_encode($link));
    }


    /**
     * Return SQL string for hostlist with count. Note, we need hostkey in order to build proper include context
     *
     * @param <CF_RULE> $rule
     * @return string
     */
    function get_SQLSTRING_HostlistAndCountAvailableUpdates($rule)
    {
        $WHERE_CONDITION = $this->_getSoftwareUpdateConditions($rule);

        $SQL = 'SELECT h.hostname AS "Host name", count(s.PatchName) AS "Updates triggering" '
                . 'FROM SoftwareUpdates s 
                INNER JOIN hosts h on h.hostkey = s.hostkey '
                . $WHERE_CONDITION . ' '
                . 'GROUP BY h.hostkey, h.hostname ORDER BY count(s.PatchName) DESC';

        return $SQL;
    }
}
?>