\set ON_ERROR_STOP true
-------------------------------------------------------------------------------
-- Custom types:
-------------------------------------------------------------------------------

DO $$
BEGIN

IF NOT EXISTS (SELECT 1 FROM pg_type WHERE typname = 'view_type') THEN
CREATE TYPE VIEW_TYPE AS ENUM ('RBAC', 'FUNCTION');
END IF;

IF NOT EXISTS (SELECT 1 FROM pg_type WHERE typname = 'change_operation') THEN
CREATE TYPE CHANGE_OPERATION AS ENUM ('ADD', 'CHANGE', 'REMOVE', 'UNTRACKED');
END IF;

IF NOT EXISTS (SELECT 1 FROM pg_type WHERE typname = 'agent_exec_status') THEN
CREATE TYPE AGENT_EXEC_STATUS AS ENUM ('OK', 'FAIL');
END IF;

IF NOT EXISTS (SELECT 1 FROM pg_type WHERE typname = 'patch_status') THEN
CREATE TYPE PATCH_STATUS AS ENUM ('INSTALLED', 'AVAILABLE');
END IF;

IF NOT EXISTS (SELECT 1 FROM pg_type WHERE typname = 'last_seen_direction') THEN
CREATE TYPE LAST_SEEN_DIRECTION AS ENUM ('INCOMING', 'OUTGOING');
END IF;

IF NOT EXISTS (SELECT 1 FROM pg_type WHERE typname = 'promise_outcome') THEN
CREATE TYPE PROMISE_OUTCOME AS ENUM ('KEPT', 'NOTKEPT', 'REPAIRED');
END IF;
END$$;

-------------------------------------------------------------------------------
-- Meta tables to handle superhub partitioned table schema
-------------------------------------------------------------------------------

DROP TABLE IF EXISTS __Views; -- want to start fresh each time
CREATE TABLE IF NOT EXISTS __Views (
       ViewType            VIEW_TYPE,
       SourceName          TEXT   PRIMARY KEY,
       ViewName            TEXT   NOT NULL,
       HostKeyColumn       TEXT,
       Id                  BIGSERIAL -- use this for ordering the views the same as in this sql file
       );
COMMENT ON TABLE __Views IS 'A list of table or function views to apply.';

-------------------------------------------------------------------------------
-- Functions and Operators:
-------------------------------------------------------------------------------

CREATE OR REPLACE FUNCTION reverse_lower_eq(text,text)
RETURNS BOOL AS
    'select lower($2) = lower($1)'
LANGUAGE SQL IMMUTABLE STRICT PARALLEL SAFE;

-------------------------------------------------------------------------------

DROP OPERATOR IF EXISTS ~~~% (text, text);
CREATE OPERATOR ~~~% (procedure = reverse_lower_eq, leftarg = text, rightarg = text);

-------------------------------------------------------------------------------

CREATE OR REPLACE FUNCTION reverse_lower_like(text,text)
RETURNS BOOL AS
    'select lower($2) LIKE lower($1)'
LANGUAGE SQL IMMUTABLE STRICT PARALLEL SAFE;

-------------------------------------------------------------------------------

DROP OPERATOR IF EXISTS ~~~%~ (text, text);
CREATE OPERATOR ~~~%~ (procedure = reverse_lower_like, leftarg = text, rightarg = text);

-------------------------------------------------------------------------------

CREATE OR REPLACE FUNCTION verify_table(in_table_name TEXT, in_table_schema TEXT DEFAULT current_schema()) RETURNS BOOL AS $$
DECLARE rec RECORD;
BEGIN
  SELECT * FROM information_schema.tables WHERE table_name = LOWER(in_table_name) AND table_schema = in_table_schema AND table_type = 'BASE TABLE' INTO rec;
  RAISE DEBUG 'verify_table(%,%) found information_schema.tables row: %', in_table_name, in_table_schema, rec;
  RETURN FOUND;
END;
$$ LANGUAGE plpgsql;

-------------------------------------------------------------------------------

CREATE OR REPLACE FUNCTION verify_schema(in_schema_name TEXT) RETURNS BOOL AS $$
DECLARE rec RECORD;
BEGIN
  SELECT * FROM information_schema.schemata WHERE schema_name = in_schema_name INTO rec;
  RAISE DEBUG 'verify_schema(%) found information_schema.schemata row: %', in_schema_name, rec;
  RETURN FOUND;
END;
$$ LANGUAGE plpgsql;

-------------------------------------------------------------------------------

CREATE OR REPLACE FUNCTION get_hub_id(hostkey_search TEXT) RETURNS NUMERIC AS $$
DECLARE out_hub_id NUMERIC;
BEGIN
  SELECT hub_id FROM __hubs where hostkey = hostkey_search INTO out_hub_id;
  IF out_hub_id IS NULL THEN
    RAISE EXCEPTION 'No hostkey of % in __hubs', hostkey_search;
  END IF;
  RAISE DEBUG 'get_hub_id(%), got hub_id=%', hostkey_search, out_hub_id;
  RETURN out_hub_id;
END;
$$ LANGUAGE plpgsql;

-------------------------------------------------------------------------------

CREATE OR REPLACE FUNCTION get_rbac_sql() RETURNS TEXT AS $$
BEGIN
  RETURN ' JOIN (
        SELECT ContextCache.HostKey
        FROM ContextCache
        WHERE ContextVector @@ to_tsquery(''simple'', current_setting(''rbac.filter''::text))
        ) AS contexts
      ON contexts.hostkey = p.hostkey; ';
END;
$$ LANGUAGE plpgsql;

-------------------------------------------------------------------------------

-- table name parameters MAY include a namespace portion like <namespace/schema>.<table_name>
-- if no namespace is included in the table name then the current_schema() will be used.
CREATE OR REPLACE FUNCTION verify_inheritance(parent_table TEXT, child_table TEXT) RETURNS BOOL AS $$
DECLARE value BOOLEAN;
        parent_info TEXT[];
        child_info TEXT[];
        parent_table_name TEXT;
        parent_namespace TEXT;
        child_table_name TEXT;
        child_namespace TEXT;
        rec RECORD;

BEGIN
  parent_info = regexp_split_to_array(parent_table, '\.');
  IF parent_info[2] IS NULL THEN
    parent_table_name = parent_info[1];
    parent_namespace = current_schema();
  ELSE
    parent_table_name = parent_info[2];
    parent_namespace = parent_info[1];
  END IF;

  child_info = regexp_split_to_array(child_table, '\.');
  IF child_info[2] IS NULL THEN
    child_table_name = child_info[1];
    child_namespace = current_schema();
  ELSE
    child_table_name = child_info[2];
    child_namespace = child_info[1];
  END IF;

  SELECT EXISTS
    (SELECT *
     FROM pg_inherits
     WHERE inhparent =
       (SELECT oid
        FROM pg_class
        WHERE relname = parent_table_name
          AND relnamespace =
           (SELECT oid
            FROM pg_namespace
            WHERE nspname = parent_namespace))
       AND inhrelid =
         (SELECT oid
          FROM pg_class
          WHERE relname = child_table_name
            AND relnamespace =
             (SELECT oid
              FROM pg_namespace
              WHERE nspname = child_namespace))) INTO value;
  RETURN value;
END;
$$ LANGUAGE plpgsql;

-------------------------------------------------------------------------------
-- Functions for cfdb database used by Mission Portal
-------------------------------------------------------------------------------


-- remove '{ ',' '} and {" "," "} from inventory slist fields
CREATE OR REPLACE FUNCTION cf_clearSlist(varchar) RETURNS varchar AS
$$ 	SELECT trim(regexp_replace(regexp_replace($1,'\'',\''|\'', \''|", ?"', ', ', 'g'), '(^\s*{\''|\''}|{"|"}\s*$)', '', 'g'))
$$ LANGUAGE SQL IMMUTABLE RETURNS NULL on NULL INPUT;

-------------------------------------------------------------------------------

-- convert string to number, return NULL if not able to convert
-- Using EXCEPTION creates subtransaction that is not PARALLEL SAFE
CREATE OR REPLACE FUNCTION cf_convertToNum(text) RETURNS numeric AS $$
    DECLARE x NUMERIC;
    BEGIN
        x = $1::NUMERIC;
        RETURN x;
    EXCEPTION WHEN others THEN
        RETURN NULL;
    END;
$$ LANGUAGE plpgsql IMMUTABLE RETURNS NULL on NULL INPUT;

-------------------------------------------------------------------------------

-- this is here so it appears just before it's first use
CREATE OR REPLACE FUNCTION is_superhub() RETURNS BOOL AS $$
BEGIN
  -- __hubs presence is our indicator of superhub or not
  RETURN public.verify_table('__hubs','public');
END;
$$ LANGUAGE plpgsql;

-------------------------------------------------------------------------------

-- return hostkeys and grouped attributes values as table by attribute name
CREATE OR REPLACE FUNCTION getHostkeyAndValuesForAttribute(attrSubStr TEXT) RETURNS TABLE (
   hostkey   text,
   inventoryValue text
  ) AS
$func$

DECLARE sqlStr TEXT;

BEGIN

sqlStr := $x$
   SELECT HostKey,
	  string_agg(level2.value, ', ' ORDER BY lower(level2.value) ASC )
	  FROM
		   (SELECT
			inventory.HostKey,
			inventory.value
		   FROM
		      inventory
		   WHERE
			 $x$|| quote_nullable(attrSubStr)  ||$x$  = ANY (inventory.metatags)
		   GROUP BY
			inventory.hostkey,
			inventory.value) as level2
	GROUP BY HostKey;
    $x$;
RETURN query
EXECUTE sqlStr
USING   attrSubStr;

  end;  $func$ language plpgsql PARALLEL SAFE;

-------------------------------------------------------------------------------

-- return hostkeys and grouped attributes values as table by key name
CREATE OR REPLACE FUNCTION getHostkeyAndValuesForKeyName(keyName TEXT) RETURNS TABLE (
   hostkey   text,
   inventoryValue text
  ) AS
$func$

DECLARE sqlStr TEXT;

BEGIN

sqlStr := $x$
   SELECT HostKey,
	  string_agg(level2.value, ', ' ORDER BY lower(level2.value) ASC )
	  FROM
		   (SELECT
			inventory.HostKey,
			inventory.value
		   FROM
		      inventory
		   WHERE
			 keyname = $x$  || quote_nullable(keyName) || $x$
		   GROUP BY
			inventory.hostkey,
			inventory.value) as level2
	GROUP BY HostKey;
    $x$;
RETURN query
EXECUTE sqlStr
USING   keyName;

  end;  $func$ language plpgsql PARALLEL SAFE;

-------------------------------------------------------------------------------

-- generate partition name for promiselog
CREATE OR REPLACE FUNCTION promise_log_partition_name_create(date TIMESTAMPTZ, outcome PROMISE_OUTCOME)
RETURNS TEXT  AS
$BODY$
DECLARE
    name TEXT;

BEGIN
    name := '__promiselog_' || outcome || '_' || to_char(date, 'YYYY-MM-DD');
    RETURN name;
END;
$BODY$
LANGUAGE plpgsql;

-------------------------------------------------------------------------------

-- Auto paritioning function for promise_log.
-- Not generic at this point as it creates indexing as well
CREATE OR REPLACE FUNCTION promise_log_partition_function()
RETURNS TRIGGER AS
$BODY$
DECLARE
    _tablename text;
    _startdate text;
    _enddate text;
    _outcome public.PROMISE_OUTCOME;

BEGIN
    -- Generate partition name based on ChangeTimeStamp and PromiseOutcome
    _outcome := NEW."promiseoutcome";
    _tablename := public.promise_log_partition_name_create(NEW."changetimestamp"::TIMESTAMPTZ, _outcome::public.PROMISE_OUTCOME);

    -- Insert the current record into the correct partition, which we are sure will now exist.
    EXECUTE 'INSERT INTO ' || TG_TABLE_SCHEMA || '.' || quote_ident(_tablename) || ' VALUES ($1.*)' USING NEW;
    RETURN NULL;

    EXCEPTION
    WHEN undefined_table THEN

    PERFORM public.promise_log_partition_create(NEW."changetimestamp", 1, _outcome, TG_TABLE_SCHEMA);
    EXECUTE 'INSERT INTO ' || TG_TABLE_SCHEMA || '.' || quote_ident(_tablename) || ' VALUES ($1.*)' USING NEW;

    RETURN NULL;
END;
$BODY$
LANGUAGE plpgsql;

-------------------------------------------------------------------------------

DROP FUNCTION IF EXISTS promise_log_partition_create(TIMESTAMPTZ, INTEGER, PROMISE_OUTCOME);
DROP FUNCTION IF EXISTS promise_log_partition_create(TIMESTAMPTZ, INTEGER, PROMISE_OUTCOME, TEXT);
CREATE OR REPLACE FUNCTION promise_log_partition_create(now TIMESTAMPTZ, partitions INTEGER, outcome PROMISE_OUTCOME, _schema TEXT DEFAULT current_schema())
RETURNS VOID AS
$BODY$

DECLARE
    _interval INTERVAL;
    _now TIMESTAMPTZ;
    _table_time TIMESTAMPTZ;
    _tablename TEXT;

BEGIN
    _interval := quote_literal(partitions - 1|| ' day');
    _now := to_timestamp(to_char(now, 'YYYY-MM-DD'), 'YYYY-MM-DD');

    FOR _table_time IN
        SELECT ts::timestamp
            FROM generate_series(_now::timestamp, _now::timestamp + _interval, '1 day'::INTERVAL) as ts
    LOOP
        _tablename := public.promise_log_partition_name_create(_table_time, outcome);

        IF NOT EXISTS (
            SELECT 1
            FROM   information_schema.tables
            WHERE  table_name = _tablename
            AND table_schema = _schema )
        THEN
            -- create new partition
            EXECUTE 'CREATE TABLE ' || quote_ident(_schema) || '.' || quote_ident(_tablename) || ' (
                CHECK ( "changetimestamp" >= ' || quote_literal(_table_time) || '::timestamptz
                AND "changetimestamp" < ' || quote_literal(_table_time + INTERVAL '1 day') || '::timestamptz
                AND "promiseoutcome" = ' || quote_literal(outcome) || '::public.PROMISE_OUTCOME )
                ) INHERITS (' || quote_ident(_schema) || '.__PromiseLog)';

            -- Indexes are defined per child, so we assign a default index that uses the partition columns
            EXECUTE 'CREATE INDEX IF NOT EXISTS ' || quote_ident(_tablename||'_hostkey') || ' ON ' || quote_ident(_tablename) || ' (HostKey) WITH (FILLFACTOR = 90)';
            PERFORM public.ensure_unique_index(quote_ident(_tablename||'_id'), quote_ident(_tablename), 'Id', 90);
            EXECUTE 'CREATE INDEX IF NOT EXISTS ' || quote_ident(_tablename||'_outcome') || ' ON ' || quote_ident(_tablename) || ' (PromiseOutcome) WITH (FILLFACTOR = 90)';
            EXECUTE 'CREATE INDEX IF NOT EXISTS ' || quote_ident(_tablename||'_changetimestamp') || ' ON ' || quote_ident(_tablename) || ' (ChangeTimeStamp) WITH (FILLFACTOR = 90)';
            EXECUTE 'CREATE INDEX IF NOT EXISTS ' || quote_ident(_tablename||'_policyfile') || ' ON ' || quote_ident(_tablename) || ' (lower(PolicyFile)) WITH (FILLFACTOR = 90)';
            EXECUTE 'CREATE INDEX IF NOT EXISTS ' || quote_ident(_tablename||'_promisehash') || ' ON ' || quote_ident(_tablename) || ' (lower(PromiseHash)) WITH (FILLFACTOR = 90)';
            EXECUTE 'CREATE INDEX IF NOT EXISTS ' || quote_ident(_tablename||'_bundlename') || ' ON ' || quote_ident(_tablename) || ' (lower(BundleName)) WITH (FILLFACTOR = 90)';
            EXECUTE 'CREATE INDEX IF NOT EXISTS ' || quote_ident(_tablename||'_promisetype') || ' ON ' || quote_ident(_tablename) || ' (lower(PromiseType)) WITH (FILLFACTOR = 90)';
            EXECUTE 'CREATE INDEX IF NOT EXISTS ' || quote_ident(_tablename||'_promiser') || ' ON ' || quote_ident(_tablename) || ' (lower(Promiser)) WITH (FILLFACTOR = 90)';
            EXECUTE 'CREATE INDEX IF NOT EXISTS ' || quote_ident(_tablename||'_stackpath') || ' ON ' || quote_ident(_tablename) || ' (lower(StackPath)) WITH (FILLFACTOR = 90)';
            EXECUTE 'CREATE INDEX IF NOT EXISTS ' || quote_ident(_tablename||'_promisehandle') || ' ON ' || quote_ident(_tablename) || ' (lower(PromiseHandle)) WITH (FILLFACTOR = 90)';

        END IF;
    END LOOP;

    RETURN;
END;
$BODY$
LANGUAGE plpgsql;

-------------------------------------------------------------------------------

-- Cleanup promiselog based on interval
CREATE OR REPLACE FUNCTION promise_log_partition_cleanup(outcome PROMISE_OUTCOME, keep_days INTERVAL)
RETURNS VOID AS
$BODY$
DECLARE
    _tablename text;

BEGIN
    FOR _tablename IN
        SELECT table_name
        FROM (
            SELECT table_name, to_timestamp(substring(table_name from (15 + char_length(outcome::text)) for 10), 'YYYY MM DD') as timestamp
            FROM information_schema.tables
            WHERE table_schema='public'
                AND table_type='BASE TABLE'
                AND table_name ILIKE '__promiselog_' || outcome || '_%'
        ) as sub
        WHERE timestamp < NOW() - keep_days
    LOOP
        EXECUTE 'DROP TABLE ' || quote_ident(_tablename);
    END LOOP;

    RETURN;
END;
$BODY$
LANGUAGE plpgsql;

-------------------------------------------------------------------------------

-- Create function to add new custom variables to variables_dictionary
CREATE OR REPLACE FUNCTION update_variables_dictionary()
    RETURNS void AS $$
BEGIN
  IF pg_is_in_recovery() THEN
       RAISE LOG 'PostgreSQL is in recovery, skipping update of variables dictionary';
  ELSIF pg_is_in_backup() THEN
    RAISE LOG 'PostgreSQL is in backup, skipping update of variables dictionary';
  ELSE
    INSERT INTO variables_dictionary ("attribute_name", "category","type","readonly","keyname")
        SELECT distinct on (attribute_name) "substring"(__variables.metatags::text, 'attribute_name=["]?([^"},]+)'::text) AS attribute_name,
        'User defined'::text as category,
        variabletype as type,
        '0'::int as readonly,
        comp as keyname
        FROM __variables
        WHERE  ('inventory'::text = ANY (__variables.metatags)) AND NOT ('attribute_name=none'::text = ANY (__variables.metatags))
        AND "substring"(__variables.metatags::text, 'attribute_name=["]?([^"},]+)'::text) NOT IN (SELECT attribute_name FROM variables_dictionary )
        GROUP BY __variables.metatags , variabletype, comp;

    INSERT INTO variables_dictionary ("attribute_name", "category","type","readonly","keyname")
        SELECT distinct on (attribute_name) "substring"(__contexts.metatags::text, 'attribute_name=["]?([^"},]+)'::text) AS attribute_name,
        'User defined'::text as category,
        'string'::text as type,
        '0'::int as readonly,
        ''::text as keyname
        FROM __contexts
        WHERE  ('inventory'::text = ANY (__contexts.metatags)) AND NOT ('attribute_name=none'::text = ANY (__contexts.metatags))
        AND "substring"(__contexts.metatags::text, 'attribute_name=["]?([^"},]+)'::text) NOT IN (SELECT attribute_name FROM variables_dictionary )
        GROUP BY __contexts.metatags;
  END IF;
END;
$$ LANGUAGE plpgsql;

-------------------------------------------------------------------------------

-- Disable variable which don't have any reports data
CREATE OR REPLACE FUNCTION disable_unreported_variables()
  RETURNS void AS $$
BEGIN
  IF pg_is_in_recovery() THEN
    RAISE LOG 'PostgreSQL is in recovery, skipping disabling unreported variables';
  ELSIF pg_is_in_backup() THEN
    RAISE LOG 'PostgreSQL is in backup, skipping disabling unreported variables';
  ELSE
    UPDATE variables_dictionary SET enabled = 0;
    UPDATE variables_dictionary SET enabled = 1 WHERE attribute_name IN ( SELECT distinct on (attribute_name) "substring"(__variables.metatags::text, 'attribute_name=["]?([^"},]+)'::text) AS attribute_name  FROM __variables) OR attribute_name IN ( SELECT distinct on (attribute_name) "substring"(__contexts.metatags::text, 'attribute_name=["]?([^"},]+)'::text) AS attribute_name  FROM __contexts);
  END IF;
END;
$$ LANGUAGE plpgsql;

-------------------------------------------------------------------------------

-- Create function  update_materialized_hosts_view
CREATE OR REPLACE FUNCTION update_materialized_hosts_view()
    RETURNS void AS $$
BEGIN
  IF pg_is_in_recovery() THEN
    RAISE LOG 'PostgreSQL is in recovery, skipping refresh of materialized view m_hosts';
  ELSIF pg_is_in_backup() THEN
    RAISE LOG 'PostgreSQL is in backup, skipping refresh of materialized view m_hosts';
  ELSE
    RAISE NOTICE 'Refreshing materialized view m_hosts';
    REFRESH MATERIALIZED VIEW CONCURRENTLY m_hosts;
  END if;
END;
$$ LANGUAGE plpgsql;

-------------------------------------------------------------------------------

CREATE OR REPLACE FUNCTION ensure_view(view_type VIEW_TYPE, source_name TEXT, view_name TEXT default '', hostkey_column TEXT default 'HostKey') RETURNS VOID AS $body$
BEGIN
  -- special general rule for our schema, private tables are prefixed with '__' and have
  -- rbac views on top without the prefix.
  EXECUTE 'INSERT INTO __Views VALUES ($1, $2, $3, $4) ON CONFLICT DO NOTHING' USING view_type, source_name, view_name, hostkey_column;

  IF view_type = 'RBAC' THEN
    IF view_name = '' THEN
      view_name := replace(source_name, '__', '');
    END IF;

    EXECUTE format('
CREATE OR REPLACE VIEW %s WITH (security_barrier) AS
  SELECT p.*
  FROM %s p JOIN
    (SELECT HostKey
     FROM ContextCache
     WHERE ((current_setting(''rbac.filter''::text) = '''') OR (ContextVector @@ to_tsquery(''simple'', current_setting(''rbac.filter''::text))))) AS contexts
  ON contexts.hostkey = p.%s;',
       view_name,
       source_name,
       hostkey_column);
  ELSIF view_type = 'FUNCTION' THEN
    EXECUTE format('SELECT %s();', source_name);
  END IF;
END;
$body$ LANGUAGE plpgsql;

-------------------------------------------------------------------------------

CREATE OR REPLACE FUNCTION ensure_primary_key(
    table_name TEXT,
    key_name TEXT,
    keys TEXT) RETURNS void AS
$BODY$
DECLARE
  key_text TEXT;
  target_schema TEXT default 'public';
BEGIN
  key_text := keys;

  IF public.is_superhub() THEN
    key_text := key_text || ', hub_id';
    target_schema := 'hub_0';
  END IF;

  IF NOT EXISTS (SELECT constraint_name
                   FROM information_schema.table_constraints
                  WHERE constraint_name = key_name
                    AND table_schema = target_schema) THEN
    EXECUTE format('ALTER TABLE %s.%s
                      ADD CONSTRAINT %s PRIMARY KEY (%s);',
                   target_schema,key_name,
                   key_name,key_text);
  END IF;
END;
$BODY$ LANGUAGE plpgsql STRICT;

COMMENT ON FUNCTION ensure_primary_key (text, text, text) IS 'create primary key in proper schema if not already present';

CREATE OR REPLACE FUNCTION ensure_unique_index(
    index_name TEXT,
    table_name TEXT,
    keys TEXT default 'hostkey', -- comma separated keys for index
    fill_factor INTEGER default 90) RETURNS void AS
$BODY$
BEGIN
    PERFORM public.ensure_index(index_name, table_name, keys, fill_factor, 'UNIQUE');
END;
$BODY$ LANGUAGE plpgsql STRICT;

COMMENT ON FUNCTION ensure_unique_index (text, text, text, integer) IS 'create unique index in proper schema';

CREATE OR REPLACE FUNCTION ensure_index(
    index_name TEXT,
    table_name TEXT,
    keys TEXT default 'hostkey', -- comma separated keys for index
    fill_factor INTEGER default 90,
    unique_text TEXT default 'UNIQUE') RETURNS void AS
$BODY$
DECLARE
  key_text TEXT;
BEGIN
  key_text := keys;
  IF public.is_superhub() THEN
    key_text := key_text || ', hub_id';
  END IF;

  EXECUTE format('
CREATE %s INDEX IF NOT EXISTS %s ON %s (%s) WITH (FILLFACTOR = %s);',
    unique_text,
    index_name,
    table_name,
    key_text,
    fill_factor);
END;
$BODY$ LANGUAGE plpgsql STRICT;

COMMENT ON FUNCTION ensure_index (text, text, text, integer, text) IS 'create index in proper schema';

-------------------------------------------------------------------------------

-- Create function  clean_historical_data
CREATE OR REPLACE FUNCTION cleanup_historical_data(
    table_name TEXT,
    order_by_column TEXT,
    items_number_to_keep INTEGER,
    hostkey_column TEXT default 'HostKey'
    ) RETURNS void AS
$BODY$
DECLARE
    row record;
BEGIN
    table_name := $1;
    order_by_column := $2;
    items_number_to_keep := $3;
    hostkey_column := $4;

    FOR row IN
        EXECUTE 'SELECT DISTINCT '|| hostkey_column ||' FROM '|| table_name ||''
    LOOP
        EXECUTE 'DELETE FROM '|| table_name ||' WHERE '|| order_by_column ||' NOT IN (SELECT '|| order_by_column ||' FROM '|| table_name ||' WHERE '|| hostkey_column ||' = '''|| row.host ||''' ORDER BY '|| order_by_column ||' DESC OFFSET 0 LIMIT '|| items_number_to_keep ||') AND '|| hostkey_column ||' = '''|| row.host ||'''';
    END LOOP;

  RETURN;
END;
$BODY$ LANGUAGE plpgsql STRICT;

COMMENT ON FUNCTION cleanup_historical_data (text, text, integer, text) IS 'Clean up stale records from table. Arguments: table name, order by column, number items to preserve, host key column name.';

-------------------------------------------------------------------------------
-- RBAC
-------------------------------------------------------------------------------

CREATE TABLE IF NOT EXISTS ContextCache (
        HostKey             TEXT        NOT NULL,
        ContextVector       TSVECTOR    NOT NULL
    ) WITH (FILLFACTOR = 90);

SELECT ensure_unique_index('contextcache_uniq', 'contextcache');
CREATE INDEX IF NOT EXISTS context_cache_vector ON ContextCache USING gin(ContextVector);
COMMENT ON TABLE ContextCache IS 'Classes reported by each host presented as a vector for full text search. Maintenance of the table is shared between triggers and cf-hub';

-------------------------------------------------------------------------------
-- Current state tables:
-------------------------------------------------------------------------------

CREATE TABLE IF NOT EXISTS __Hosts (
        HostKey                 TEXT                        NOT NULL,
        IsCallCollected         boolean                     DEFAULT false,
        LastReportTimeStamp     timestamp with time zone    DEFAULT CURRENT_TIMESTAMP,
--      HostName                TEXT,                       comment: virtual field in hosts view mapped from variables
--      IPAddress               TEXT,                       comment: virtual field in hosts view mapped from lastseenhosts
        FirstReportTimeStamp    timestamp with time zone    DEFAULT CURRENT_TIMESTAMP,
        HostKeyCollisions       INTEGER                     DEFAULT 0,
        CONSTRAINT hostkey_not_empty_check CHECK(hostkey <> '')
    ) WITH (FILLFACTOR = 70);

ALTER TABLE __Hosts ADD COLUMN IF NOT EXISTS HostKeyCollisions INTEGER DEFAULT 0;
DO $$
  BEGIN
    ALTER TABLE "__hosts" DROP CONSTRAINT IF EXISTS hostkey_not_empty_check;
    ALTER TABLE "__hosts" ADD CONSTRAINT hostkey_not_empty_check CHECK(hostkey <> '');
    EXCEPTION
      WHEN check_violation THEN
        raise NOTICE 'Failed to add constraint hostkey_not_empty_check to __hosts table, there are entries that have empty hostkey. Please contact support@northern.tech';
  END;
$$
language plpgsql;

COMMENT ON TABLE __Hosts IS 'Internal table holding basic host information.';
COMMENT ON COLUMN __Hosts.HostKey IS 'Public key digest representing host identity (unique).';
COMMENT ON COLUMN __Hosts.IsCallCollected IS 'True when report collection has been initiated by the client at least one time.';
COMMENT ON COLUMN __Hosts.LastReportTimeStamp IS 'Timestamp of last successful report collection by cf-hub (seconds precision).';
COMMENT ON COLUMN __Hosts.FirstReportTimeStamp IS 'Timestamp of when the host first appeared in hosts table, when it was discovered for the first time by cf-hub.';
COMMENT ON COLUMN __Hosts.HostKeyCollisions IS 'Number of times cf-hub observed cookie mismatch (from protocol version 3)';

SELECT ensure_unique_index('__hosts_uniq', '__hosts', 'hostkey', 70);

SELECT ensure_view('RBAC', '__Hosts', 'v_Hosts');

CREATE OR REPLACE FUNCTION ensure_hosts_not_reported_view() RETURNS VOID AS $body$
BEGIN
  CREATE OR REPLACE VIEW hosts_not_reported WITH (security_barrier) AS
    SELECT p.*
    FROM __hosts p
    WHERE (EXISTS (
            SELECT HostKey
            FROM ContextCache
            WHERE ContextVector @@ to_tsquery('simple', 'host.not.reported')
                AND p.hostkey = HostKey
        ) OR NOT EXISTS (
            SELECT HostKey
            FROM ContextCache
            WHERE p.hostkey = HostKey
        )) AND (firstreporttimestamp + INTERVAL '1 hour')::timestamptz < now();
END;
$body$ LANGUAGE plpgsql;

SELECT ensure_view('FUNCTION', 'ensure_hosts_not_reported_view');

COMMENT ON VIEW hosts_not_reported IS 'View to select never collected hosts';

-------------------------------------------------------------------------------

CREATE TABLE IF NOT EXISTS __AgentStatus (
        HostKey                     		TEXT                          NOT NULL,
        AgentExecutionInterval      		INTEGER,
        LastAgentLocalExecutionTimeStamp 	timestamp with time zone,
        LastAgentExecutionStatus    		AGENT_EXEC_STATUS
    ) WITH (FILLFACTOR = 70);

SELECT ensure_unique_index('__agentstatus_uniq', '__agentstatus', 'hostkey', 70);
SELECT ensure_view('RBAC', '__AgentStatus');

-------------------------------------------------------------------------------

CREATE TABLE IF NOT EXISTS __Contexts (
        HostKey             TEXT                        NOT NULL,
        ContextName         TEXT                        NOT NULL,
        MetaTags            TEXT[],
        ChangeTimeStamp     timestamp with time zone
    ) WITH (FILLFACTOR = 90);

SELECT ensure_unique_index('__contexts_uniq', '__contexts', 'hostkey, contextname', 90);
CREATE INDEX IF NOT EXISTS contexts_pattern_ops ON __Contexts (ContextName varchar_pattern_ops) WITH (FILLFACTOR = 90);
CREATE INDEX IF NOT EXISTS contexts_meta ON __Contexts (MetaTags) WITH (FILLFACTOR = 90);

SELECT ensure_view('RBAC', '__Contexts');

-------------------------------------------------------------------------------

CREATE TABLE IF NOT EXISTS __Variables (
        HostKey             TEXT                        NOT NULL,
        NameSpace           TEXT                        DEFAULT NULL,
        Bundle              TEXT                        NOT NULL,
        VariableName        TEXT                        NOT NULL,
        VariableValue       TEXT                        DEFAULT NULL,
        VariableType        TEXT                        NOT NULL,
        comp                TEXT                        NOT NULL,
        MetaTags            TEXT[],
        ChangeTimeStamp     timestamp with time zone
    ) WITH (FILLFACTOR = 70);

SELECT ensure_unique_index('__variables_uniq', '__variables', 'hostkey, comp', 70);
CREATE INDEX IF NOT EXISTS variables_meta ON __Variables (MetaTags) WITH (FILLFACTOR = 70);
CREATE INDEX IF NOT EXISTS variables_variablename ON __Variables (VariableName) WITH (FILLFACTOR = 70);
CREATE INDEX IF NOT EXISTS variables_variablevalue ON __Variables (VariableValue varchar_pattern_ops) WITH (FILLFACTOR = 70);

SELECT ensure_view('RBAC', '__Variables');

-------------------------------------------------------------------------------

CREATE TABLE IF NOT EXISTS __Software (
        HostKey                 TEXT                        NOT NULL,
        compo                   TEXT                        NOT NULL,
        SoftwareName            TEXT                        NOT NULL,
        SoftwareVersion         TEXT                        DEFAULT NULL,
        SoftwareArchitecture    TEXT                        NOT NULL,
        ChangeTimeStamp		    timestamp with time zone
    ) WITH (FILLFACTOR = 90);

-- ENT-7419: unknown version should be represented as NULL (but older versions of the schema forbid that)
ALTER TABLE __Software ALTER COLUMN SoftwareVersion DROP NOT NULL;

SELECT ensure_unique_index('__software_uniq', '__software', 'hostkey, compo', 90);
CREATE INDEX IF NOT EXISTS software_name_version ON __Software (SoftwareName, SoftwareVersion) with (FILLFACTOR = 90);

SELECT ensure_view('RBAC', '__Software');

-------------------------------------------------------------------------------

CREATE TABLE IF NOT EXISTS __SoftwareUpdates (
        HostKey             TEXT                        NOT NULL,
        compo               TEXT                        NOT NULL,
        PatchName           TEXT                        NOT NULL,
        PatchVersion        TEXT                        NOT NULL,
        PatchArchitecture   TEXT                        NOT NULL,
        PatchReportType     PATCH_STATUS                NOT NULL,
        ChangeTimeStamp   	timestamp with time zone
    ) WITH (FILLFACTOR = 90);

SELECT ensure_unique_index('__softwareupdates_uniq', '__softwareupdates', 'hostkey, compo', 90);
CREATE INDEX IF NOT EXISTS software_name ON __SoftwareUpdates (PatchName) WITH (FILLFACTOR = 90);

SELECT ensure_view('RBAC', '__SoftwareUpdates');

-------------------------------------------------------------------------------

CREATE TABLE IF NOT EXISTS __LastSeenHosts (
        HostKey             TEXT                        NOT NULL,
        compo               TEXT                        NOT NULL,
        LastSeenDirection   LAST_SEEN_DIRECTION         NOT NULL,
        RemoteHostKey       TEXT                        NOT NULL,
        RemoteHostIP        TEXT,
        LastSeenTimeStamp   timestamp with time zone    NOT NULL,
        LastSeenInterval    real                        NOT NULL
    ) WITH (FILLFACTOR = 70);

SELECT ensure_unique_index('__lastseenhosts_uniq', '__lastseenhosts', 'hostkey, compo', 70);
CREATE INDEX IF NOT EXISTS lastseenhosts_remotekey ON __LastSeenHosts (RemoteHostKey) WITH (FILLFACTOR = 70);
SELECT ensure_view('RBAC', '__LastSeenHosts');

----------------------Drop __lastseenhostslog table and references-------------
DROP TABLE IF EXISTS  __LastSeenHostsLogs CASCADE;
DROP FUNCTION IF EXISTS  lastseen_hosts_log_function CASCADE;
-------------------------------------------------------------------------------

CREATE TABLE IF NOT EXISTS __PromiseExecutions (
        HostKey             TEXT                        NOT NULL,
        PolicyFile          TEXT                        NOT NULL,
        ReleaseId           TEXT                        DEFAULT NULL,
        PromiseHash         TEXT                        NOT NULL,
        NameSpace           TEXT                        NOT NULL,
        BundleName          TEXT                        NOT NULL,
        PromiseType         TEXT                        NOT NULL,
        Promiser            TEXT                        DEFAULT NULL,
        StackPath           TEXT                        NOT NULL,
        PromiseHandle       TEXT                        DEFAULT NULL,
        PromiseOutcome      CHARACTER VARYING(8)        NOT NULL,
        LineNumber          INTEGER                     DEFAULT 0,
        PolicyFileHash      TEXT                        DEFAULT NULL,
        LogMessages         TEXT[]                      DEFAULT NULL,
        Promisees           TEXT[]                      DEFAULT NULL,
        ChangeTimeStamp     timestamp with time zone
    ) WITH (FILLFACTOR = 70);

-- change promiseoutcome data type to character varying
DO $$
  BEGIN
    IF NOT EXISTS (
      SELECT 1 from information_schema.columns
      WHERE table_name ILIKE '__promiseexecutions'
        AND table_schema = 'public'
        AND column_name = 'promiseoutcome'
        AND data_type = 'character varying')
      THEN
        DROP VIEW IF EXISTS promiseexecutions;
        ALTER TABLE __PromiseExecutions ALTER COLUMN promiseoutcome type CHARACTER VARYING(8) USING promiseoutcome::text;
    END IF;
END$$;

-- added later, may need to be added to an existing table on updates
ALTER TABLE __PromiseExecutions ADD COLUMN IF NOT EXISTS LineNumber INTEGER DEFAULT 0;
ALTER TABLE __PromiseExecutions ADD COLUMN IF NOT EXISTS PolicyFileHash TEXT DEFAULT NULL;

SELECT ensure_unique_index('__promiseexecutions_uniq', '__promiseexecutions', 'hostkey, promisehash', 70);
CREATE INDEX IF NOT EXISTS promiseexecutions_bundlename ON __PromiseExecutions (BundleName) WITH (FILLFACTOR = 70);
CREATE INDEX IF NOT EXISTS promiseexecutions_promiser ON __PromiseExecutions (Promiser) WITH (FILLFACTOR = 70);
CREATE INDEX IF NOT EXISTS promiseexecutions_promiseoutsome ON __PromiseExecutions (PromiseOutcome) WITH (FILLFACTOR = 70);
CREATE INDEX IF NOT EXISTS promiseexecutions_promisees ON __PromiseExecutions (Promisees) WITH (FILLFACTOR =  70);
CREATE INDEX IF NOT EXISTS promiseexecutions_promisehash ON __PromiseExecutions (PromiseHash) WITH (FILLFACTOR =  70);
SELECT ensure_view('RBAC', '__PromiseExecutions');

-------------------------------------------------------------------------------
-- Log tables:
-------------------------------------------------------------------------------

CREATE TABLE IF NOT EXISTS __ContextsLog (
        HostKey             TEXT                        NOT NULL,
        ChangeTimeStamp     timestamp with time zone    NOT NULL,
        ChangeOperation     CHANGE_OPERATION            NOT NULL,
        ContextName         TEXT                        NOT NULL,
        MetaTags            TEXT[]
    );

CREATE INDEX IF NOT EXISTS contexts_log_hostkey ON __ContextsLog (HostKey) WITH (FILLFACTOR = 70);

SELECT ensure_view('RBAC', '__ContextsLog');

-------------------------------------------------------------------------------

CREATE TABLE IF NOT EXISTS __VariablesLog (
        HostKey             TEXT                        NOT NULL,
        ChangeTimeStamp     timestamp with time zone    NOT NULL,
        ChangeOperation     CHANGE_OPERATION            NOT NULL,
        NameSpace           TEXT                        DEFAULT NULL,
        Bundle              TEXT                        NOT NULL,
        VariableName        TEXT                        NOT NULL,
        VariableValue       TEXT                        DEFAULT NULL,
        VariableType        TEXT                        NOT NULL,
        MetaTags            TEXT[]
    );

CREATE INDEX IF NOT EXISTS variables_log_hostkey ON __VariablesLog (HostKey) WITH (FILLFACTOR = 70);

SELECT ensure_view('RBAC', '__VariablesLog');

-------------------------------------------------------------------------------

CREATE TABLE IF NOT EXISTS __SoftwareLog (
        HostKey                 TEXT                        NOT NULL,
        ChangeTimeStamp         timestamp with time zone    NOT NULL,
        ChangeOperation         CHANGE_OPERATION            NOT NULL,
        SoftwareName            TEXT                        NOT NULL,
        SoftwareVersion         TEXT                        DEFAULT NULL,
        SoftwareArchitecture    TEXT                        NOT NULL
    );

-- ENT-7419: unknown version should be represented as NULL (but older versions of the schema forbid that)
ALTER TABLE __SoftwareLog ALTER COLUMN SoftwareVersion DROP NOT NULL;

CREATE INDEX IF NOT EXISTS software_log_hostkey ON __SoftwareLog (HostKey) WITH (FILLFACTOR = 70);

SELECT ensure_view('RBAC', '__SoftwareLog');

-------------------------------------------------------------------------------

CREATE TABLE IF NOT EXISTS __SoftwareUpdatesLog (
        HostKey             TEXT                        NOT NULL,
        ChangeTimeStamp     timestamp with time zone    NOT NULL,
        ChangeOperation     CHANGE_OPERATION            NOT NULL,
        PatchName           TEXT                        NOT NULL,
        PatchVersion        TEXT                        NOT NULL,
        PatchArchitecture   TEXT                        NOT NULL,
        PatchReportType     PATCH_STATUS                NOT NULL
    );

CREATE INDEX IF NOT EXISTS software_updates_log_hostkey ON __SoftwareUpdatesLog (HostKey) WITH (FILLFACTOR = 70);

SELECT ensure_view('RBAC', '__SoftwareUpdatesLog');

-------------------------------------------------------------------------------

CREATE TABLE IF NOT EXISTS __FileChangesLog (
        HostKey             TEXT                        NOT NULL,
        PromiseHandle       TEXT                        NOT NULL,
        FileName            TEXT                        NOT NULL,
        ChangeTimeStamp     timestamp with time zone    NOT NULL,
        ChangeType          TEXT                        NOT NULL,
        ChangeDetails       TEXT[]                      DEFAULT NULL
    );

CREATE INDEX IF NOT EXISTS filechanges_hk ON __FileChangesLog (HostKey) WITH (FILLFACTOR = 90);

SELECT ensure_view('RBAC', '__FileChangesLog');

-------------------------------------------------------------------------------

CREATE TABLE IF NOT EXISTS __BenchmarksLog (
        HostKey             TEXT                        NOT NULL,
        EventName           TEXT                        NOT NULL,
        StandardDeviation   NUMERIC                     NOT NULL,
        AverageValue        NUMERIC                     NOT NULL,
        LastValue           NUMERIC                     NOT NULL,
        CheckTimeStamp      timestamp with time zone    NOT NULL
    );

CREATE INDEX IF NOT EXISTS benchmarkslog_hk ON __BenchmarksLog (HostKey) WITH (FILLFACTOR = 90);

SELECT ensure_view('RBAC', '__BenchmarksLog');

-------------------------------------------------------------------------------

-- Deprecated in 3.7.0 (replaces by PromiseLog)
-- TODO: Remove in 3.8.0

CREATE TABLE IF NOT EXISTS __PromiseExecutionsLog (
        HostKey             TEXT                        NOT NULL,
        ChangeTimeStamp     timestamp with time zone    NOT NULL,
        ChangeOperation     CHANGE_OPERATION            NOT NULL,
        PolicyFile          TEXT                        NOT NULL,
        ReleaseId           TEXT                        DEFAULT NULL,
        PromiseHash         TEXT                        NOT NULL,
        NameSpace           TEXT                        NOT NULL,
        BundleName          TEXT                        NOT NULL,
        PromiseType         TEXT                        NOT NULL,
        Promiser            TEXT                        DEFAULT NULL,
        StackPath           TEXT                        NOT NULL,
        PromiseHandle       TEXT                        DEFAULT NULL,
        PromiseOutcome      PROMISE_OUTCOME             NOT NULL,
        LogMessages         TEXT[]                      DEFAULT NULL,
        Promisees           TEXT[]                      DEFAULT NULL
    );

CREATE INDEX IF NOT EXISTS promiseexecutions_log_hostkey ON __PromiseExecutionsLog (HostKey) WITH (FILLFACTOR = 70);

SELECT ensure_view('RBAC', '__PromiseExecutionsLog');

-------------------------------------------------------------------------------

CREATE TABLE IF NOT EXISTS __PromiseLog (
        Id                  bigserial                   primary key,
        HostKey             TEXT                        NOT NULL,
        ChangeTimeStamp     timestamp with time zone    NOT NULL,
        PolicyFile          TEXT                        NOT NULL,
        ReleaseId           TEXT                        DEFAULT NULL,
        PromiseHash         TEXT                        NOT NULL,
        NameSpace           TEXT                        NOT NULL,
        BundleName          TEXT                        NOT NULL,
        PromiseType         TEXT                        NOT NULL,
        Promiser            TEXT                        DEFAULT NULL,
        StackPath           TEXT                        NOT NULL,
        PromiseHandle       TEXT                        DEFAULT NULL,
        PromiseOutcome      PROMISE_OUTCOME             NOT NULL,
        LineNumber          INTEGER                     DEFAULT 0,
        PolicyFileHash      TEXT                        DEFAULT NULL,
        LogMessages         TEXT[]                      DEFAULT NULL,
        Promisees           TEXT[]                      DEFAULT NULL
    );

-- added later, may need to be added to an existing table on updates
ALTER TABLE __PromiseLog ADD COLUMN IF NOT EXISTS LineNumber INTEGER DEFAULT 0;
ALTER TABLE __PromiseLog ADD COLUMN IF NOT EXISTS PolicyFileHash TEXT DEFAULT NULL;

CREATE INDEX IF NOT EXISTS promiselog_promiseoutcome ON __PromiseLog (PromiseOutcome);
CREATE INDEX IF NOT EXISTS promiselog_promisehash ON __PromiseLog (PromiseHash);
CREATE INDEX IF NOT EXISTS promiselog_changetimestamp_promiseoutcome ON __promiselog ("changetimestamp", "promiseoutcome");

-- Drop view to change column type to bigint --
DROP VIEW IF EXISTS PromiseLog;
-- Must drop dependent view before altering __PromiseLog
DROP VIEW IF EXISTS not_kept_not_repaired;
ALTER TABLE __PromiseLog ALTER COLUMN id type bigint;

SELECT ensure_view('RBAC', '__PromiseLog');

DROP TRIGGER IF EXISTS promise_log_partition_trigger ON __PromiseLog;
CREATE TRIGGER promise_log_partition_trigger
BEFORE INSERT ON __PromiseLog
FOR EACH ROW EXECUTE PROCEDURE promise_log_partition_function();

-- Create empty partitions during installation for 7 days in the past and 3 days up in the future.
SELECT promise_log_partition_create(NOW() - INTERVAL '7 day', 7 + 3, 'REPAIRED');
SELECT promise_log_partition_create(NOW() - INTERVAL '7 day', 7 + 3, 'NOTKEPT');

-------------------------------------------------------------------------------
-- Monitoring tables:
-------------------------------------------------------------------------------

-- 3.6.2 changes schema for __MonitoringMg, during upgrade old one has to be removed
DO $$
DECLARE
TableExists boolean;
BEGIN
    SELECT INTO TableExists 1 FROM information_schema.tables WHERE table_name = lower('__MonitoringMgMeta');
    IF NOT FOUND THEN
        DROP TABLE IF EXISTS __MonitoringMg;
    END IF;
END
$$ ;


CREATE TABLE IF NOT EXISTS __MonitoringMgMeta (
        id                  bigserial   PRIMARY KEY,
        hostkey             TEXT        NOT NULL,
        observable          TEXT        NOT NULL,
        global              BOOLEAN,
        expected_min        REAL,
        expected_max        REAL,
        unit                TEXT,
        description         TEXT,
        updatedTimeStamp    timestamp with time zone,
        lastUpdatedSample   INTEGER
) WITH (FILLFACTOR = 70);
COMMENT ON TABLE __MonitoringMgMeta IS 'Stores 1 record for ech observable per host.';

-- Drop view to change column type to bigint --
DROP VIEW IF EXISTS MonitoringMgMeta;
ALTER TABLE __MonitoringMgMeta ALTER COLUMN id type bigint;
SELECT ensure_unique_index('__MonitoringMgMeta_uniq', '__MonitoringMgMeta', 'hostkey, observable', 70);
SELECT ensure_view('RBAC', '__MonitoringMgMeta');

CREATE TABLE IF NOT EXISTS __MonitoringMg (
        meta_id                bigserial     REFERENCES __MonitoringMgMeta (id) ON DELETE CASCADE,
        sample                 INTEGER,
        value1                 REAL,
        value2                 REAL,
        value3                 REAL,
        value4                 REAL,
        updatedTimeStamp       timestamp with time zone
) WITH (FILLFACTOR = 70);
COMMENT ON TABLE __MonitoringMg IS 'Stores average daily values for each observable over a week. Stores 2016 records per observable per host.';

ALTER TABLE __MonitoringMg ALTER COLUMN meta_id type bigint;
SELECT ensure_unique_index('__MonitoringMg_uniq', '__MonitoringMg', 'meta_id, sample', 70);
CREATE INDEX IF NOT EXISTS monitoring_mg_time ON __MonitoringMg (updatedTimeStamp) WITH (FILLFACTOR = 70);

-------------------------------------------------------------------------------

-- 3.6.2 changes schema for __MonitoringYr, during upgrade old one has to be removed
DO $$
DECLARE
TableExists boolean;
BEGIN
    SELECT INTO TableExists 1 FROM information_schema.tables WHERE table_name = lower('__MonitoringYrMeta');
    IF NOT FOUND THEN
        DROP TABLE IF EXISTS __MonitoringYr;
    END IF;
END
$$ ;


CREATE TABLE IF NOT EXISTS __MonitoringYrMeta (
        id                  bigserial   PRIMARY KEY,
        hostkey             TEXT        NOT NULL,
        observable          TEXT        NOT NULL,
        global              BOOLEAN,
        expected_min        REAL,
        expected_max        REAL,
        unit                TEXT,
        description         TEXT,
        lastUpdatedSample   INTEGER
) WITH (FILLFACTOR = 70);
COMMENT ON TABLE __MonitoringYrMeta is 'Stores 1 record for each observable per host';

-- Drop view to change column type to bigint --
DROP VIEW IF EXISTS MonitoringYrMeta;
ALTER TABLE __MonitoringYrMeta ALTER COLUMN id type bigint;

SELECT ensure_unique_index('__MonitoringYrMeta_uniq', '__MonitoringYrMeta', 'hostkey, observable', 70);

SELECT ensure_view('RBAC', '__MonitoringYrMeta');


CREATE TABLE IF NOT EXISTS __MonitoringYr (
        meta_id                bigserial     REFERENCES __MonitoringYrMeta (id) ON DELETE CASCADE,
        sample                 INTEGER,
        value1                 REAL,
        value2                 REAL,
        value3                 REAL,
        value4                 REAL
) WITH (FILLFACTOR = 70);
COMMENT ON TABLE __MonitoringYr IS 'Stores average weekly values for each observable over 3 years. Stores 156 records per observable per host.';

ALTER TABLE __MonitoringYr ALTER COLUMN meta_id type bigint;

SELECT ensure_unique_index('__MonitoringYr_uniq', '__MonitoringYr', 'meta_id, sample', 70);

-------------------------------------------------------------------------------

CREATE TABLE IF NOT EXISTS __MonitoringHg (
        host    TEXT        NOT NULL,
        id      TEXT        NOT NULL,
        ar1     REAL[]
    ) WITH (FILLFACTOR = 70);
COMMENT ON TABLE __MonitoringHg IS 'Stores 1 record for each observable per host.';

SELECT ensure_unique_index('__MonitoringHg_uniq', '__MonitoringHg', 'host, id', 70);

SELECT ensure_view('RBAC', '__MonitoringHg', 'MonitoringHg', 'host');

-------------------------------------------------------------------------------
-- Diagnostics tables:
-------------------------------------------------------------------------------

CREATE TABLE IF NOT EXISTS __HubConnectionErrors (
        CheckTimeStamp      timestamp with time zone    DEFAULT CURRENT_TIMESTAMP,
        HostKey             TEXT                        NOT NULL,
        Message             TEXT                        NOT NULL,
        QueryType           TEXT                        NOT NULL
    );

SELECT ensure_view('RBAC', '__HubConnectionErrors');

-------------------------------------------------------------------------------

-- 3.6.2 Renames '__Diagnostics' to 'Diagnostics' table
ALTER TABLE IF EXISTS __Diagnostics RENAME TO Diagnostics;


CREATE TABLE IF NOT EXISTS Diagnostics (
        Name           TEXT                         NOT NULL,
        Details        TEXT,
        TimeStamp      timestamp with time zone     DEFAULT CURRENT_TIMESTAMP,
        Value          NUMERIC                      NOT NULL,
        Units          TEXT                         NOT NULL
    );

-- ENT-4331: 3.14.0/3.12.2 renames 'Status' to '__Status' and creates a RBAC-checked view
-- ENT-5396: Upgrades from 3.12.1 to 3.12.2 and 3.12.3 leave us with both __status and status tables.
-- we must ensure things are correct and merge data if need be.
DO $$
BEGIN
  IF verify_table('__status') AND verify_table('status') THEN
    RAISE NOTICE 'Found both status and __status tables. Migrating all data to __status to make proper.';
    CREATE TABLE tmp_status (LIKE __status);
    INSERT INTO tmp_status SELECT * FROM __status;
    INSERT INTO tmp_status SELECT * FROM status;
    DROP TABLE __status;
    DROP TABLE status;
    ALTER TABLE tmp_status RENAME TO __status;
  ELSIF verify_table('status') THEN
    ALTER TABLE status RENAME to __status;
  END IF;
END $$;
    
CREATE TABLE IF NOT EXISTS __STATUS (
        host           CHARACTER VARYING(256)   NOT NULL,
        ts             CHARACTER VARYING(256)   NOT NULL,
        hubts          timestamp with time zone NOT NULL DEFAULT CURRENT_TIMESTAMP,
        status         CHARACTER VARYING(256),
        lstatus        CHARACTER VARYING(256),
        type           CHARACTER,
        who            INTEGER,
        whr            INTEGER,
        PRIMARY        KEY(host,ts)
    ) WITH (FILLFACTOR = 90);

ALTER TABLE __STATUS
ADD COLUMN IF NOT EXISTS hubts timestamp with time zone NOT NULL DEFAULT CURRENT_TIMESTAMP;

COMMENT ON TABLE __STATUS IS 'Report collection status';
COMMENT ON COLUMN __STATUS.ts IS 'Timestamp of last data provided by client during report collection. This is used by delta queries to request a start time.';

SELECT ensure_view('RBAC', '__Status', 'Status', 'host');

-------------------------------------------------------------------------------

-- Update ContextCache after all tables are created (Useful during hub upgrade)
-- Warning: order is important.

TRUNCATE TABLE ContextCache;

INSERT INTO ContextCache (hostkey, contextvector)
    SELECT hostkey, to_tsvector('simple', replace(x::text,'_','.'))
    FROM (
        SELECT hostkey, array_agg(contextname) as x
        FROM __Contexts
        GROUP BY hostkey
        ) as sub;

INSERT INTO ContextCache(hostkey, contextvector)
    SELECT hostkey, to_tsvector('simple',replace('host_not_reported'::text,'_','.'))
    FROM __hosts
    WHERE hostkey NOT IN (SELECT hostkey FROM ContextCache);

----------------------------------------------------
--- ENT-4554 Change m_inventory into a regular table
----------------------------------------------------

-- Drop old inventory views and function --
DO $$
DECLARE
  record_count integer;
BEGIN

  SELECT count(*)
  INTO record_count
  FROM pg_matviews
  WHERE matviewname = 'inventory_new';

  IF record_count = 1 then
    EXECUTE 'DROP MATERIALIZED VIEW public.inventory_new';
  ELSE
    EXECUTE 'DROP VIEW IF EXISTS public.inventory_new';
  END IF;

END $$;

DROP MATERIALIZED VIEW IF EXISTS public.m_inventory;
DROP FUNCTION IF EXISTS update_materialized_inventory_view;

CREATE TABLE IF NOT EXISTS  __inventory (
    hostkey text PRIMARY KEY,
    values jsonb NOT NULL
) WITH (FILLFACTOR = 90);

COMMENT ON TABLE "__inventory" IS 'Stores normalized inventory data';

SELECT ensure_view('RBAC', '__inventory', 'inventory_new');

COMMENT ON VIEW "inventory_new" IS 'Regular view based on __inventory table with applied RBAC filter';

CREATE OR REPLACE FUNCTION update_inventory()
    RETURNS void AS $$
BEGIN
  IF pg_is_in_recovery() THEN
    RAISE LOG 'PostgreSQL is in recovery, skipping update of __inventory table';
  ELSIF pg_is_in_backup() THEN
    RAISE LOG 'PostgreSQL is in backup, skipping update of __inventory table';
  ELSIF public.is_superhub() THEN
    RAISE DEBUG 'superhub is enabled, __inventory will not be updated';
  ELSE
    RAISE NOTICE 'Updating __inventory table';
        TRUNCATE __inventory;
        INSERT INTO __inventory (SELECT * FROM v_inventory WHERE values IS NOT NULL);
  END IF;
END;
$$ LANGUAGE plpgsql;
COMMENT ON FUNCTION update_inventory IS 'Update __inventory table';

CREATE OR REPLACE FUNCTION ensure_v_inventory_view() RETURNS VOID AS $body$
DECLARE hosts_hub_id_column TEXT;
BEGIN
  IF public.is_superhub() THEN
    hosts_hub_id_column = ', __hosts.hub_id ';
  END IF;

  EXECUTE format('
CREATE OR REPLACE VIEW public.v_inventory AS
SELECT __hosts.hostkey,
( SELECT json_object(array_agg(d.keyname), array_agg(d.value)) AS array_to_json
  FROM (
    SELECT  keyname,
    string_agg(cf_clearslist(value),'', '') AS value
    FROM( SELECT substring(__variables.metatags::TEXT FROM ''attribute_name=["]?([^"},]+)'') AS keyname,
      string_agg(cf_clearslist(__variables.variablevalue),'', '')   AS value
      FROM __variables
      WHERE __variables.hostkey = __hosts.hostkey
      AND (''inventory''::text = ANY (__variables.metatags)) AND (__variables.metatags::text LIKE ''%%attribute_name%%'') AND NOT (''attribute_name=none''::text = ANY (__variables.metatags))
      GROUP BY __variables.metatags

      UNION ALL

      SELECT substring(__contexts.metatags::TEXT FROM ''attribute_name=["]?([^"},]+)'') AS keyname,
      string_agg(cf_clearslist(__contexts.contextname),'', '') AS value
      FROM __contexts
      WHERE __contexts.hostkey = __hosts.hostkey
      AND (''inventory''::text = ANY (__contexts.metatags)) AND (__contexts.metatags::text LIKE ''%%attribute_name%%'') AND NOT (''attribute_name=none''::text = ANY (__contexts.metatags))
      GROUP BY __contexts.metatags)f  GROUP BY keyname
    )d
  ) AS "values"
%s -- __hosts.hub_id
FROM __hosts;', hosts_hub_id_column);

EXECUTE 'SELECT update_inventory()'; -- because we may have just added hub_id column

COMMENT ON VIEW "v_inventory" IS 'Regular view to get normalized inventory data';
END;
$body$ LANGUAGE plpgsql;

SELECT ensure_view('FUNCTION', 'ensure_v_inventory_view');

CREATE OR REPLACE FUNCTION update_inventory_by_hostkey(hostkey TEXT)
  RETURNS void AS $$
BEGIN
  IF pg_is_in_recovery() THEN
    RAISE LOG 'PostgreSQL is in recovery, skipping update of __inventory table';
  ELSIF pg_is_in_backup() THEN
    RAISE LOG 'PostgreSQL is in backup, skipping update of __inventory table';
  ELSE
    RAISE NOTICE 'Updating __inventory table';
    EXECUTE 'DELETE FROM __inventory WHERE hostkey = ''' || hostkey || '''';
    EXECUTE 'INSERT INTO __inventory (SELECT * FROM parse_inventory_by_hostkey('''|| hostkey ||''') WHERE hostkey IS NOT NULL)';
  END IF;
END;
$$ LANGUAGE plpgsql;

COMMENT ON FUNCTION update_inventory_by_hostkey IS 'Update specific host inventory data';

DROP FUNCTION IF EXISTS parse_inventory_by_hostkey(TEXT);
CREATE OR REPLACE FUNCTION parse_inventory_by_hostkey(hostkey TEXT)
  RETURNS __inventory AS $$
DECLARE
  row __inventory%rowtype;
  inventoryViewDefinition TEXT;
  hostExists int;
BEGIN
EXECUTE 'SELECT 1 FROM __hosts WHERE hostkey = '''|| hostkey ||'''' INTO hostExists;
  IF hostExists = 1 THEN
      EXECUTE 'SELECT pg_get_viewdef(''v_inventory'', true)' INTO inventoryViewDefinition;
      inventoryViewDefinition = RTRIM(inventoryViewDefinition, ';');
      EXECUTE inventoryViewDefinition || ' WHERE __hosts.hostkey = '''|| hostkey ||'''' INTO row;
      IF row.values IS NULL THEN
          -- if there is no inventory json values then set empty object --
         row.values = '{}';
      END IF;
      RETURN row;
  ELSE
      RAISE NOTICE 'Host % not found. Notice raised while parse_inventory_by_hostkey function performing', hostkey;
      RETURN NULL;
  END IF;
END;
$$ LANGUAGE plpgsql;

COMMENT ON FUNCTION parse_inventory_by_hostkey IS 'Parse specific host inventory data from __variables and __contexts';

-------------------------------------------------------------------------------
-- Hosts view and m_hosts materialized view
-------------------------------------------------------------------------------

-- Create Hosts and m_hosts views
CREATE OR REPLACE FUNCTION ensure_hosts_views() RETURNS VOID AS $body$
DECLARE hosts_hub_id_text TEXT;
        and_inventory_hub_id_text TEXT;
        cte_hub_id_text TEXT;
        lastseen_hub_id_text TEXT;
        and_lastseen_hub_id_text TEXT;
BEGIN
  IF public.is_superhub() THEN
    hosts_hub_id_text = ', __hosts.hub_id';
    and_inventory_hub_id_text = ' AND inv.hub_id=__hosts.hub_id';
    cte_hub_id_text = ', CTE.hub_id';
    lastseen_hub_id_text = ', __LastSeenHosts.hub_id';
    and_lastseen_hub_id_text = ' AND __hosts.hub_id=l.hub_id';
  END IF;
  EXECUTE 'DROP MATERIALIZED VIEW IF EXISTS m_hosts CASCADE';
  EXECUTE format('
CREATE MATERIALIZED VIEW m_hosts AS
(SELECT
        __hosts.hostkey,
        inv.values->>''Host name'' as hostname,
        l.ipaddress,
        __hosts.lastreporttimestamp,
        __hosts.firstreporttimestamp,
        __hosts.hostkeycollisions
        %s
FROM __hosts
LEFT JOIN __inventory AS inv ON inv.hostkey = __hosts.hostkey %s
LEFT JOIN (
    SELECT
        HostKey,
        IPAddress
        %s
    FROM (
        SELECT
            RemoteHostKey AS HostKey,
            RemoteHostIP AS IPAddress,
            ROW_NUMBER() OVER (PARTITION BY RemoteHostKey)
            %s
        FROM __LastSeenHosts
    ) AS CTE
    WHERE
        row_number=1
) l
ON __hosts.HostKey=l.HostKey %s
);', hosts_hub_id_text, and_inventory_hub_id_text, cte_hub_id_text, lastseen_hub_id_text, and_lastseen_hub_id_text);
  ALTER TABLE public.m_hosts OWNER TO cfapache;
END;
$body$ LANGUAGE plpgsql;

SELECT ensure_view('FUNCTION', 'ensure_hosts_views');
SELECT ensure_view('RBAC', 'm_hosts', 'hosts');
SELECT ensure_unique_index('m_hosts_uniq', 'm_hosts');

-- Refresh materialized view in case of a host was bootstrapped
CREATE OR REPLACE FUNCTION refresh_m_hosts_function()
RETURNS TRIGGER AS
$BODY$
BEGIN
  IF pg_is_in_recovery() THEN
    RAISE LOG 'PostgreSQL is in recovery, skipping bootstrap triggered refresh of materialized view m_hosts';
  ELSIF pg_is_in_backup() THEN
    RAISE LOG 'PostgreSQL is in backup, skipping bootstrap triggered refresh of materialized view m_hosts';
  ELSE
    RAISE NOTICE 'Host bootstrap triggering refreshing m_hosts';
    REFRESH MATERIALIZED VIEW CONCURRENTLY m_hosts;
  END IF;
  RETURN NULL;
END;
$BODY$
LANGUAGE plpgsql;

-- Manage hosts_insert_trigger and hosts_delete_trigger
DO $$
DECLARE
  target_schema TEXT default 'public';
BEGIN
  IF public.is_superhub() THEN
    target_schema := 'hub_0';
  END IF;

  EXECUTE format('
DROP TRIGGER IF EXISTS hosts_insert_trigger ON %s.__hosts;
CREATE TRIGGER hosts_insert_trigger
AFTER INSERT ON %s.__hosts
FOR EACH ROW EXECUTE PROCEDURE refresh_m_hosts_function();

DROP TRIGGER IF EXISTS hosts_delete_trigger ON %s.__hosts;
CREATE TRIGGER hosts_delete_trigger
AFTER DELETE ON %s.__hosts
FOR EACH STATEMENT EXECUTE PROCEDURE refresh_m_hosts_function();
',
  target_schema, target_schema,
  target_schema, target_schema);
END$$;


-- Clear hosts references in case of a host was deleted
CREATE OR REPLACE FUNCTION clear_hosts_references()
RETURNS TRIGGER AS
$BODY$
BEGIN
    DELETE FROM contextcache WHERE hostkey = OLD.hostkey;
    DELETE FROM __agentstatus WHERE hostkey = OLD.hostkey;
    DELETE FROM __benchmarkslog WHERE hostkey = OLD.hostkey;
    DELETE FROM __contexts WHERE hostkey = OLD.hostkey;
    DELETE FROM __contextslog WHERE hostkey = OLD.hostkey;
    DELETE FROM __filechangeslog WHERE hostkey = OLD.hostkey;
    DELETE FROM __health_diagnostics_dismissed WHERE hostkey = OLD.hostkey;
    DELETE FROM __hubconnectionerrors WHERE hostkey = OLD.hostkey;
    DELETE FROM __inventory WHERE hostkey = OLD.hostkey;
    DELETE FROM __lastseenhosts WHERE hostkey = OLD.hostkey;
    DELETE FROM __monitoringyrmeta WHERE hostkey = OLD.hostkey;
    DELETE FROM __monitoringmgmeta WHERE hostkey = OLD.hostkey;
    DELETE FROM __monitoringhg WHERE host = OLD.hostkey;
    DELETE FROM __promiseexecutions WHERE hostkey = OLD.hostkey;
    DELETE FROM __promiseexecutionslog WHERE hostkey = OLD.hostkey;
    DELETE FROM __promiselog WHERE hostkey = OLD.hostkey;
    DELETE FROM __software WHERE hostkey = OLD.hostkey;
    DELETE FROM __softwarelog WHERE hostkey = OLD.hostkey;
    DELETE FROM __softwareupdates WHERE hostkey = OLD.hostkey;
    DELETE FROM __softwareupdateslog WHERE hostkey = OLD.hostkey;
    DELETE FROM __status WHERE host = OLD.hostkey;
    DELETE FROM __variables WHERE hostkey = OLD.hostkey;
    DELETE FROM __variableslog WHERE hostkey = OLD.hostkey;
    RETURN NULL;
END;
$BODY$
LANGUAGE plpgsql;

DO $$
DECLARE
  target_schema TEXT default 'public';
BEGIN
  IF public.is_superhub() THEN
    target_schema := 'hub_0';
  END IF;

  EXECUTE format('
DROP TRIGGER IF EXISTS hosts_clear_refrences_trigger ON %s.__hosts;
CREATE TRIGGER hosts_clear_refrences_trigger
AFTER DELETE ON %s.__hosts
FOR EACH ROW EXECUTE PROCEDURE clear_hosts_references();
',
  target_schema, target_schema);
END$$;

-- DROP VIEW public.v_contextcache
DROP VIEW IF EXISTS public.v_contextcache;

-- Create TABLE public.variables_dictionary
CREATE TABLE IF NOT EXISTS public.variables_dictionary
(
  id serial   PRIMARY KEY,
  attribute_name character varying(200),
  category character varying(200),
  readonly integer DEFAULT 0,
  type character varying(200),
  convert_function character varying(200),
  keyname character varying(200),
  CONSTRAINT variables_dictionary_attribute_name UNIQUE (attribute_name)
)
WITH (
  OIDS=FALSE
);

ALTER TABLE variables_dictionary
ADD COLUMN IF NOT EXISTS enabled integer DEFAULT 0;

ALTER TABLE public.variables_dictionary OWNER TO cfpostgres; -- squashed to make unit testing easier
GRANT SELECT ON TABLE public.variables_dictionary TO public;
GRANT ALL ON TABLE public.variables_dictionary TO cfpostgres;


-------------------------------------------------------------------------------
-- Events table :
-------------------------------------------------------------------------------

CREATE TABLE IF NOT EXISTS  events (
    id        bigserial   PRIMARY KEY,
    event character varying(30) NOT NULL,
    payload json DEFAULT '{}' NOT NULL,
    date timestamptz DEFAULT CURRENT_TIMESTAMP NOT NULL
) WITH (FILLFACTOR = 90);

-------------------------------------------------------------------------------
-- Function and trigger to handle decommissioned and bootstrapped hosts events :
-------------------------------------------------------------------------------

CREATE OR REPLACE FUNCTION process_host_events() RETURNS TRIGGER AS $host_log_event_trigger$
BEGIN
  IF pg_is_in_recovery() THEN
    RAISE LOG 'PostgreSQL is in recovery, skipping processing host events';
  ELSIF pg_is_in_backup() THEN
    RAISE LOG 'PostgreSQL is in backup, skipping processing host events';
  ELSE
    RAISE NOTICE 'Processing host events';
    IF (TG_OP = 'DELETE') THEN
      INSERT INTO events  (event, payload) VALUES ('host_decommissioned', json_build_object('hostkey', OLD.hostkey));
      RETURN OLD;
    ELSIF (TG_OP = 'INSERT') THEN
      INSERT INTO events  (event, payload) VALUES ('host_bootstrapped', json_build_object('hostkey', NEW.hostkey));
      RETURN NEW;
    END IF;
  END IF;
  RETURN NULL;
END;
$host_log_event_trigger$ LANGUAGE plpgsql;

-- Manage host_log_event_trigger
DO $$
DECLARE
  target_schema TEXT default 'public';
BEGIN
  IF public.is_superhub() THEN
    target_schema := 'hub_0';
  END IF;

  EXECUTE format('
DROP TRIGGER IF EXISTS host_log_event_trigger ON %s.__hosts;
CREATE TRIGGER host_log_event_trigger
AFTER INSERT OR DELETE ON %s.__hosts
    FOR EACH ROW EXECUTE PROCEDURE process_host_events();
',
  target_schema, target_schema);
END$$;

---- ENT-3950 ----

CREATE TABLE IF NOT EXISTS "__health_diagnostics_dismissed" (
    hostkey text,
    report_type character varying(30),
    username character varying(20),
    CONSTRAINT "health_diagnostics_dismissed_hostkey_report_type_username" PRIMARY KEY ("hostkey", "report_type", "username")
);

COMMENT ON TABLE "__health_diagnostics_dismissed" IS 'Table that contains specific dismissed hosts from Health diagnostic reports';

SELECT ensure_view('RBAC', '__health_diagnostics_dismissed', 'health_diagnostics_dismissed', 'hostkey');

---- ENT-4567 Add __hub table to store hub's information ----
CREATE TABLE IF NOT EXISTS __hub (
    "hostkey" text PRIMARY KEY,
    "hostname" text,
    "ip" text,
    "update_ts" timestamptz
);

COMMENT ON TABLE "__hub" IS 'Table stores hub''s information';

-- Manage privileges on events_id_seq and events table
DO $$
DECLARE
  target_schema TEXT default 'public';
BEGIN
  IF public.is_superhub() THEN
    target_schema := 'hub_0';
  END IF;

  EXECUTE format('
GRANT ALL PRIVILEGES ON SEQUENCE %s.events_id_seq TO public;
GRANT ALL PRIVILEGES ON TABLE %s.events TO public;
',
  target_schema, target_schema);
END$$;

---- ENT-4611 schema changes on enabling superhub ----

CREATE OR REPLACE FUNCTION ensure_views() RETURNS VOID AS $body$
DECLARE r record;
BEGIN
  FOR r in SELECT * FROM __Views ORDER BY Id
  LOOP
    EXECUTE 'SELECT ensure_view($1,$2,$3,$4)' USING r.viewtype, r.sourcename, r.viewname, r.hostkeycolumn;
  END LOOP;
END;
$body$ LANGUAGE plpgsql;

CREATE OR REPLACE function switch_to_feeder_schema(hostkey_search TEXT) RETURNS VOID AS $$
DECLARE id NUMERIC;
        feeder_schema TEXT;
BEGIN
  id := get_hub_id(hostkey_search);
  feeder_schema = 'hub_' || id || '_importing';

  RAISE DEBUG 'set schema %', feeder_schema;
  EXECUTE 'SET SCHEMA ' || quote_literal(feeder_schema);
END;
$$ LANGUAGE plpgsql;

CREATE OR REPLACE function obtain_shared_schema_lock() RETURNS VOID AS $$
BEGIN
  PERFORM public.obtain_schema_lock(true);
END;
$$ LANGUAGE plpgsql;

CREATE OR REPLACE function obtain_exclusive_schema_lock() RETURNS VOID AS $$
BEGIN
  PERFORM public.obtain_schema_lock(false);
END;
$$ LANGUAGE plpgsql;

CREATE OR REPLACE function obtain_schema_lock(shared BOOLEAN = false) RETURNS VOID AS $$
DECLARE tries NUMERIC := 0;
        max_tries NUMERIC := 10;
        lock_acquired BOOLEAN := false;
        lock_type TEXT;
BEGIN
  WHILE NOT lock_acquired AND tries < max_tries LOOP
    IF shared THEN
      SELECT * FROM pg_try_advisory_lock_shared(18002) INTO lock_acquired;
    ELSE
      SELECT * FROM pg_try_advisory_lock(18002) INTO lock_acquired;
    END IF;
    tries := tries + 1;
    IF NOT lock_acquired THEN
      PERFORM pg_sleep(1);
    END IF;
  END LOOP;
  IF shared THEN
    lock_type := 'shared';
  ELSE
    lock_type := 'exclusive';
  END IF;
  IF lock_acquired THEN
    RAISE DEBUG 'obtained % schema lock 18002 in % tries', lock_type, tries;
  ELSE
    RAISE EXCEPTION 'unable to obtain % schema lock 18002 in % tries', lock_type, tries;
  END IF;
END;
$$ LANGUAGE plpgsql;

CREATE OR REPLACE function release_shared_schema_lock() RETURNS VOID AS $$
BEGIN
  PERFORM public.release_schema_lock(true);
END;
$$ LANGUAGE plpgsql;

CREATE OR REPLACE function release_exclusive_schema_lock() RETURNS VOID AS $$
BEGIN
  PERFORM public.release_schema_lock(false);
END;
$$ LANGUAGE plpgsql;

CREATE OR REPLACE function release_schema_lock(shared BOOLEAN = false) RETURNS VOID AS $$
DECLARE lock_type TEXT;
BEGIN
  IF shared THEN
    lock_type := 'shared';
    PERFORM pg_advisory_unlock_shared(18002);
  ELSE
    lock_type := 'exclusive';
    PERFORM pg_advisory_unlock(18002);
  END IF;
  RAISE DEBUG 'released % schema lock 18002', lock_type;
END;
$$ LANGUAGE plpgsql;


CREATE OR REPLACE function drop_feeder_schema(hostkey_search TEXT) RETURNS VOID AS $$
DECLARE id NUMERIC;
        feeder_schema TEXT;
BEGIN
  PERFORM public.obtain_exclusive_schema_lock();

  id := get_hub_id(hostkey_search);
  feeder_schema = 'hub_' || id || '_importing';

  RAISE DEBUG 'dropping schema %', feeder_schema;
  EXECUTE 'DROP SCHEMA ' || feeder_schema || ' CASCADE';

  PERFORM public.release_exclusive_schema_lock();
EXCEPTION WHEN others THEN
  PERFORM public.release_exclusive_schema_lock();
  RAISE NOTICE 'exception in drop_feeder_schema % - %', sqlstate, sqlerrm;
END;
$$ LANGUAGE plpgsql;

CREATE OR REPLACE function ensure_feeder_schema(hostkey_search TEXT, table_whitelist TEXT[]) RETURNS VOID AS $$
DECLARE id NUMERIC;
        _tablename TEXT;
        feeder_table TEXT;
        new_schema TEXT;
BEGIN
  PERFORM public.obtain_exclusive_schema_lock();

  id := get_hub_id(hostkey_search);
  new_schema = 'hub_' || id || '_importing';

  -- renaming the schema will not affect use of it, everything should still work as before
  IF verify_schema(new_schema) THEN
    EXECUTE 'DROP SCHEMA ' || new_schema || ' CASCADE'; -- get rid of old import schema, shouldn't happen
  END IF;

  -- create the new schema in which to import feeder data
  EXECUTE 'CREATE SCHEMA ' || new_schema;
  EXECUTE 'SET SCHEMA ' || quote_literal(new_schema);

  FOREACH _tablename IN ARRAY table_whitelist
  LOOP
    RAISE DEBUG 'handling whitelist entry %', _tablename;
    -- new_schema __promiselog is a temporary table to handle import
    -- will be removed and have child tables re-attached to public.__promiselog
    -- in attach_feeder_schema().
    IF _tablename = '__promiselog_*' THEN
      _tablename := '__promiselog';
    END IF;

    feeder_table := new_schema || '.' || _tablename;

    RAISE DEBUG 'create table % like public.%', feeder_table, _tablename;
    EXECUTE 'CREATE TABLE ' || feeder_table || ' (LIKE public.' || _tablename || ' INCLUDING DEFAULTS)';

    RAISE DEBUG 'setting default hub_id of table % to %', feeder_table, id;
    EXECUTE 'ALTER TABLE ' || feeder_table || ' ALTER COLUMN hub_id SET DEFAULT ' || id;

    IF _tablename = '__promiselog' THEN
      RAISE DEBUG 'adding feeder schema trigger for __promiselog';
      EXECUTE 'CREATE TRIGGER feeder_promise_log_partition_trigger BEFORE INSERT ON ' || feeder_table || ' FOR EACH ROW EXECUTE PROCEDURE public.promise_log_partition_function()';
    END IF;

  END LOOP;

  RAISE DEBUG 'set schema %', new_schema;
  EXECUTE 'SET SCHEMA ''' || new_schema || '''';

  PERFORM public.release_exclusive_schema_lock();
  RAISE DEBUG 'done with ensure_feeder_schema, no exceptions';
EXCEPTION WHEN others THEN
  PERFORM public.release_exclusive_schema_lock();
  RAISE NOTICE 'ensure_feeder_schema had an exception % - %', sqlstate, sqlerrm;
END;
$$ LANGUAGE plpgsql;



-------------------------------------------------------------------------------
-- handle_duplicate_hostkeys_in_import()
-------------------------------------------------------------------------------
-- Provided an array of hostkeys for the current set of importing feeders
-- find duplicate hostkeys in hub_<n> and hub_<n>_importing schemas.
-- Choose the entry with the newest __hosts.lastreporttimestamp and use
-- that set of data (either public or one of the feeders schemas).
-- Move all other non-logging table data to a dup schema for later analysis.
-- Ensure that any records in dup are unique and updated each run.
CREATE OR REPLACE FUNCTION handle_duplicate_hostkeys_in_import(importing_feeder_hostkeys TEXT[]) RETURNS VOID AS $$
DECLARE tmp_schema TEXT;
        tmp_table TEXT;
        dup_schema TEXT;
        dup_table TEXT;
        source_schema TEXT; -- for moving duplicates to either importing or normal feeder schemas
        tables TEXT[] = ARRAY[
            'contextcache',
            '__agentstatus',
            '__contexts',
            '__hosts',
            '__inventory',
            '__lastseenhosts',
            '__promiseexecutions',
            '__software',
            '__softwareupdates',
            '__variables'
        ];
        schemas_union_sql TEXT = '';
        _table TEXT;
        rec RECORD;
        dup_count int := 0;
        feeder_hub_id BIGINT;
        importing_hub_ids BIGINT[];
        not_importing_hub_ids BIGINT[];
        error_code TEXT;
        error_message TEXT;
        error_stack TEXT;
BEGIN
  PERFORM public.obtain_exclusive_schema_lock();

  RAISE DEBUG '% Begin handling duplicate hostkeys in import', clock_timestamp();
  dup_schema = 'dup';
  tmp_schema = 'tmp';

  RAISE DEBUG 'Ensuring temporary and duplicate schemas exist';
  IF verify_schema(tmp_schema) THEN
    EXECUTE 'DROP SCHEMA ' || tmp_schema || ' CASCADE';
  END IF;
  EXECUTE 'CREATE SCHEMA ' || tmp_schema;
  IF NOT verify_schema(dup_schema) THEN
    EXECUTE 'CREATE SCHEMA ' || dup_schema;
  END IF;

  RAISE DEBUG 'Determining if feeders are importing or not';
  FOR rec in SELECT hostkey, hub_id from __hubs WHERE hub_id != 0
  LOOP
    IF importing_feeder_hostkeys @> ARRAY[rec.hostkey] THEN
      RAISE DEBUG 'Adding feeder %, hub_id % to importing_hub_ids', rec.hostkey, rec.hub_id;
      importing_hub_ids := array_append(importing_hub_ids, rec.hub_id);
    ELSE
      IF verify_schema('hub_' || rec.hub_id) THEN
        schemas_union_sql := schemas_union_sql || ' UNION ALL SELECT * FROM hub_' || rec.hub_id || '.__hosts ';
      END IF;
      RAISE DEBUG 'Adding feeder %, hub_id % to NOT importing hub ids', rec.hostkey, rec.hub_id;
      not_importing_hub_ids := array_append(not_importing_hub_ids, rec.hub_id);
    END IF;
  END LOOP;

  RAISE DEBUG 'Create tables in temporary and duplicate schemas';
  FOREACH _table in ARRAY tables
  LOOP
    tmp_table := tmp_schema || '.' || _table;
    dup_table := dup_schema || '.' || _table;
    EXECUTE 'CREATE TABLE ' || tmp_table || ' (LIKE public.' || _table || ' INCLUDING DEFAULTS) PARTITION BY LIST (hub_id)';
    IF NOT verify_table(_table, dup_schema) THEN
      EXECUTE 'CREATE TABLE ' || dup_table || ' (LIKE public.' || _table || ' INCLUDING DEFAULTS)';
    END IF;

    RAISE DEBUG 'Attaching importing tables to % schema', tmp_schema;
    FOREACH feeder_hub_id in ARRAY importing_hub_ids
    LOOP
      EXECUTE 'ALTER TABLE ' || tmp_table || ' ATTACH PARTITION hub_' || feeder_hub_id || '_importing.' || _table || ' FOR VALUES IN (' || feeder_hub_id || ')';
    END LOOP;
  END LOOP;

  RAISE DEBUG '% Find duplicates and move to appropriate schema', clock_timestamp();
  tmp_table := tmp_schema || '.__hosts';
  EXECUTE '
    CREATE TEMP TABLE duplicate_hostkeys AS
    SELECT
      hostkey,
      hub_id
    FROM (
      SELECT *,row_number()
        OVER (
          PARTITION BY hostkey
          ORDER BY lastreporttimestamp DESC
        ) as rnum
      FROM (
        SELECT * FROM ' || tmp_table || schemas_union_sql || '
      ) cte
    ) cte2
  WHERE cte2.rnum > 1';

  SELECT count(*) FROM duplicate_hostkeys INTO dup_count;
  RAISE INFO 'Found % duplicates. All duplicate data has been moved to schema "%" for analysis.', dup_count, dup_schema;

  FOR rec IN SELECT DISTINCT(hub_id) as hub_id FROM duplicate_hostkeys
  LOOP
    RAISE DEBUG '% Processing duplicates for hub_id %', clock_timestamp(), rec.hub_id;
    EXECUTE 'CREATE TEMP TABLE duplicates_' || rec.hub_id || ' AS SELECT hostkey FROM duplicate_hostkeys WHERE hub_id = ' || rec.hub_id;
    EXECUTE 'CREATE TEMP TABLE deletes_' || rec.hub_id || ' AS SELECT d1.hostkey FROM dup.__hosts as d1, duplicates_' || rec.hub_id || ' AS d2 WHERE d1.hostkey = d2.hostkey';

    IF ARRAY[rec.hub_id] <@ importing_hub_ids THEN
      source_schema := 'hub_' || rec.hub_id || '_importing';
    ELSE
      source_schema := 'hub_' || rec.hub_id;
    END IF;

    FOREACH _table in ARRAY tables
    LOOP
      RAISE DEBUG '% delete in dup schema for hub_id % table %', clock_timestamp(), rec.hub_id, _table;
      EXECUTE 'DELETE FROM dup.' || _table || ' WHERE hostkey IN (SELECT * FROM deletes_' || rec.hub_id || ')';
      IF _table = '__lastseenhosts'::text THEN
        EXECUTE 'DELETE FROM dup.' || _table || ' WHERE remotehostkey IN (SELECT * FROM deletes_' || rec.hub_id || ')';
      END IF;
      RAISE DEBUG '% insert from % schema to dup for hub_id % table %', clock_timestamp(), source_schema, rec.hub_id, _table;
      EXECUTE 'INSERT INTO dup.' || _table || ' SELECT * FROM ' || source_schema || '.' || _table || ' WHERE hostkey in (SELECT * FROM duplicates_' || rec.hub_id || ')';
      RAISE DEBUG '% delete from % schema for hub_id % table %', clock_timestamp(), source_schema, rec.hub_id, _table;
      EXECUTE 'DELETE FROM ' || source_schema || '.' || _table || ' WHERE hostkey in (SELECT * FROM duplicates_' || rec.hub_id || ')';
    END LOOP;
  END LOOP;

  RAISE DEBUG 'Detaching importing tables from % schema', tmp_schema;
  FOREACH _table in ARRAY tables
  LOOP
    FOREACH feeder_hub_id in ARRAY importing_hub_ids
    LOOP
      EXECUTE 'ALTER TABLE ' || tmp_schema || '.' || _table || ' DETACH PARTITION hub_' || feeder_hub_id || '_importing.' || _table;
    END LOOP;
  END LOOP;

  RAISE DEBUG 'Deleting % schema', tmp_schema;
  EXECUTE 'DROP SCHEMA ' || tmp_schema || ' CASCADE';

  PERFORM public.release_exclusive_schema_lock();

  RAISE DEBUG '% Done handling duplicate hostkeys in import', clock_timestamp();
EXCEPTION
WHEN others then
  RAISE NOTICE 'exception during handle_duplicate_hostkeys_in_import(), cleaning up';

  IF verify_schema(tmp_schema) THEN
    EXECUTE 'DROP SCHEMA ' || tmp_schema || ' CASCADE';
  END IF;
  GET STACKED DIAGNOSTICS error_code = RETURNED_SQLSTATE,
                          error_message = MESSAGE_TEXT,
                          error_stack = PG_EXCEPTION_CONTEXT;

  PERFORM public.release_exclusive_schema_lock();

  RAISE EXCEPTION 'handle_duplicate_hostkeys_in_import() got an exception: RETURNED_SQLSTATE: %, MESSAGE_TEXT: %, PG_EXCEPTION_CONTEXT: %', error_code, error_message, error_stack;
END;
$$ LANGUAGE plpgsql;

COMMENT ON FUNCTION handle_duplicate_hostkeys_in_import IS 'Given an array of hostkeys during Federated Reporting import, move duplicates based on __hosts.lastreporttimestamp to a dup schema not used by Mission Portal.';


CREATE OR REPLACE FUNCTION attach_feeder_schema(hostkey_search TEXT, table_whitelist TEXT[]) RETURNS VOID AS $$
DECLARE r RECORD;
        _tablename TEXT;
        orig_schema TEXT;
        new_schema TEXT;
        feeder_table TEXT;
        old_table TEXT;
        public_table TEXT;
        id NUMERIC;
        error_code TEXT;
        error_message TEXT;
        error_stack TEXT;
BEGIN
  PERFORM public.obtain_exclusive_schema_lock();
  RAISE DEBUG 'ENTER attach_feeder_schema(%)', hostkey_search;
  id := get_hub_id(hostkey_search);
  orig_schema := 'hub_' || id;
  new_schema := 'hub_' || id || '_importing'; -- this is the schema in which the current import has been processed

  FOREACH _tablename IN ARRAY table_whitelist
  LOOP
    IF (_tablename != '__promiselog_*') THEN
      feeder_table := quote_ident(new_schema) || '.' || quote_ident(_tablename);
      old_table := quote_ident(orig_schema) || '.' || quote_ident(_tablename);
      public_table := quote_ident('public') || '.' || quote_ident(_tablename);

      RAISE DEBUG 'feeder_table: %', feeder_table;
      RAISE DEBUG 'old_table: %', old_table;
      RAISE DEBUG 'public_table: %', public_table;

      IF verify_inheritance(public_table, old_table) THEN
        EXECUTE 'ALTER TABLE ' || public_table || ' DETACH PARTITION ' || old_table;
      END IF;

      RAISE DEBUG 'attaching new table % to public table % for values in %', feeder_table, public_table, id;
      EXECUTE 'ALTER TABLE ' || public_table || ' ATTACH PARTITION ' || feeder_table || ' FOR VALUES IN (' || id || ')';
    ELSE
      RAISE DEBUG 'handle __promiselog child tables ';
      FOR r in SELECT table_name
               FROM information_schema.tables
               WHERE table_name ~ '__promiselog_'
                 AND table_schema = new_schema
      LOOP
        feeder_table := quote_ident(new_schema) || '.' || quote_ident(r.table_name);
        RAISE DEBUG 'setting hub_id in table % to %', feeder_table, id;
        EXECUTE 'UPDATE ' || feeder_table || ' SET hub_id = ' || id;
        EXECUTE 'ALTER TABLE ' || feeder_table || ' NO INHERIT ' || new_schema || '.__promiselog';
        EXECUTE 'ALTER TABLE ' || feeder_table || ' INHERIT public.__promiselog';
      END LOOP;

      -- remove temporary feeder __promiselog table now that all child tables
      -- are re-attached to public.__promiselog.
      EXECUTE 'DROP TABLE ' || new_schema || '.__promiselog CASCADE';

    END IF;
  END LOOP;

  IF verify_schema(orig_schema) THEN
    RAISE DEBUG 'schema attachment success, remove orig schema: %', orig_schema;
    EXECUTE 'DROP SCHEMA ' || orig_schema || ' CASCADE';
  END IF;

  EXECUTE 'ALTER SCHEMA ' || new_schema || ' RENAME TO ' || orig_schema;

  PERFORM public.release_exclusive_schema_lock();
EXCEPTION
WHEN others THEN
  RAISE NOTICE 'got exception during attach_feeder_schema(), cleaning up';

  IF verify_schema(new_schema) THEN
    RAISE DEBUG 'Dropping new schema % which failed to complete', new_schema;
    EXECUTE 'DROP SCHEMA ' || new_schema || ' CASCADE';
  END IF;

  GET STACKED DIAGNOSTICS error_code = RETURNED_SQLSTATE,
                          error_message = MESSAGE_TEXT,
                          error_stack = PG_EXCEPTION_CONTEXT;

  PERFORM public.release_exclusive_schema_lock();

  RAISE EXCEPTION 'attach_feeder_schema(%) got an exception: RETURNED_SQLSTATE: %, MESSAGE_TEXT: %, PG_EXCEPTION_CONTEXT: %', hostkey_search, error_code, error_message, error_stack;
END;
$$ LANGUAGE plpgsql;

-- superhub_schema() returns exit_code=<exit_code> for parsing by wrapper script
-- exit_code = 0 (kept), 1 (repaired), 2 (failed)
DROP FUNCTION IF EXISTS superhub_schema(); -- to allow changes to function signature
DROP FUNCTION IF EXISTS superhub_schema(TEXT);
CREATE OR REPLACE FUNCTION superhub_schema(superhub_hostkey TEXT) RETURNS TEXT AS $superhub_schema$
DECLARE r record;
        hub_table TEXT;
        public_table TEXT;
        pk_name TEXT;
        exit_code INT = 2; -- failed (not kept)
        error_message TEXT;
        error_stack TEXT;
        indexes_cursor REFCURSOR;
        index_rec RECORD;
        index_sql TEXT;
        skip_tables TEXT[] = ARRAY[
          '__hubs',
          '__hub',
          '__health_diagnostics_dismissed',
          '__views',
          'variables_dictionary' -- because it is generated from partitioned __inventory
        ];

BEGIN
  IF NOT public.is_superhub() THEN
    BEGIN
      PERFORM public.obtain_exclusive_schema_lock();

      -- note that any changes to __hubs should be synchronized in mission-portal/tests/phpunit/models/RemoteHubModelTest.php
      CREATE TABLE __hubs (hub_id BIGSERIAL, hostkey TEXT PRIMARY KEY, last_import_ts TIMESTAMPTZ);
      INSERT INTO __hubs (hub_id, hostkey) VALUES (0, superhub_hostkey);

      CREATE SCHEMA hub_0;

      FOR r in SELECT table_name FROM information_schema.tables
               WHERE table_type = 'BASE TABLE' AND table_schema = 'public'
      LOOP
        hub_table := quote_ident('hub_0') || '.' || quote_ident(r.table_name);
        public_table := quote_ident(r.table_name);

        IF (NOT ARRAY[public_table] <@ skip_tables) THEN
          RAISE DEBUG 'REPAIRING table %', public_table;

          RAISE DEBUG 'adding hub_id to table %', public_table;
          EXECUTE 'ALTER TABLE ' || public_table || ' ADD COLUMN IF NOT EXISTS hub_id BIGINT DEFAULT 0';

          IF (public_table !~ '__promiselog*') THEN
            -- first get all indexes on the public table and keep them for later
            OPEN indexes_cursor FOR SELECT indexdef FROM pg_indexes
              WHERE tablename = public_table AND schemaname = 'public';

            -- must move the public table to hub_0 because tables can't be altered to be partitioned
            RAISE DEBUG 'moving table % from public to hub_0', public_table;
            EXECUTE 'ALTER TABLE ' || public_table || ' SET SCHEMA "hub_0"';

            RAISE DEBUG 'creating public partitioned table and attaching hub table to it';
            EXECUTE 'CREATE TABLE ' || public_table || ' (like ' || hub_table || ' INCLUDING DEFAULTS) PARTITION BY LIST (hub_id)';
            EXECUTE 'ALTER TABLE ' || public_table || ' ATTACH PARTITION ' || hub_table || ' FOR VALUES IN (0)';

            -- recreate the indexes on the public partitioned table with the addition of hub_id column
            -- since feeder schemas' tables will attach they will inherit all indexes on the public tables
            LOOP
              FETCH indexes_cursor INTO index_rec;
              EXIT WHEN NOT FOUND;

              index_sql := index_rec.indexdef;
              IF position('UNIQUE' in index_sql) != 0 THEN
                index_sql := regexp_replace(index_sql,'\((.*?)\)','(\1, hub_id)');
              END IF;

              index_sql := replace(index_sql,'hub_0','public');
              BEGIN
                EXECUTE index_sql;
              EXCEPTION
              WHEN others THEN
                RAISE WARNING 'failed to create index "%s"', index_sql;
              END;
            END LOOP;

            CLOSE indexes_cursor;
          END IF;
        END IF;
      END LOOP;

      EXECUTE 'SELECT ensure_views()';

      -- materialized views are not included in information_schema.tables or information_schema.views
      -- so handle m_hosts special because it is the only one.
      EXECUTE 'CREATE UNIQUE INDEX IF NOT EXISTS m_hosts_uniq on m_hosts (hostkey, hub_id) WITH (FILLFACTOR = 90)';

      -- add hub_id column as part of primary key in __promiselog
      ALTER TABLE __PromiseLog DROP CONSTRAINT IF EXISTS __promiselog_pkey;
      ALTER TABLE __PromiseLog ADD PRIMARY KEY (id, hub_id);

      PERFORM public.release_exclusive_schema_lock();

      exit_code = 1; -- repaired

    EXCEPTION
      WHEN others THEN
        GET STACKED DIAGNOSTICS error_message = MESSAGE_TEXT,
                                error_stack = PG_EXCEPTION_CONTEXT;
        RAISE NOTICE E'schema modifications failed: %\n%', error_message, error_stack;

        PERFORM public.release_exclusive_schema_lock();

        exit_code = 2; -- not kept
    END;
  ELSE -- __hubs table exists
    exit_code = 0; -- kept
  END IF;

  RETURN 'exit_code=' || exit_code;
END;
$superhub_schema$ LANGUAGE plpgsql;

-------------------------------------------------------------------------------

-- Virtual View
--CREATE TABLE IF NOT EXISTS Inventory (
--        HostKey             TEXT,
--        KeyName             TEXT,
--        Type                TEXT,
--        MetaTags            TEXT[],
--        Value               TEXT
--    );
CREATE OR REPLACE FUNCTION ensure_inventory_view() RETURNS VOID AS $body$
DECLARE hub_id_text TEXT;
BEGIN
  IF public.is_superhub() THEN
    hub_id_text = ', hub_id';
  END IF;

  EXECUTE format('
CREATE OR REPLACE VIEW Inventory WITH (security_barrier) AS (
    SELECT
        HostKey,
        KeyName,
        Type,
        MetaTags,
        Value
        %s
    FROM (
        SELECT
            HostKey,
            ContextName AS KeyName,
            ''context'' AS Type,
            MetaTags,
            ContextName AS Value
            %s
        FROM
            Contexts
        WHERE
            ''inventory'' = ANY (MetaTags)
            AND NOT ''attribute_name=none'' = ANY(MetaTags)
        UNION ALL
        SELECT
            HostKey,
            comp AS KeyName,
            VariableType AS Type,
            MetaTags,
            VariableValue AS Value
            %s
        FROM
            Variables
        WHERE
            ''inventory'' = ANY (MetaTags)
            AND NOT ''attribute_name=none'' = ANY(MetaTags)
    ) tmp
);
', hub_id_text, hub_id_text, hub_id_text);
END;
$body$ LANGUAGE plpgsql;
SELECT ensure_view('FUNCTION', 'ensure_inventory_view');
COMMENT ON VIEW Inventory is 'This is old inventory view that contains fetched data from __variables and __contexts. This view is no longer used by API or Mission portal and exists only for backward compatibility';

---- ENT-4612 ensure provided feeders are added to __hubs table ----
-- prints to stderr exit_code=<exit_code> for parsing by wrapper script
-- exit_code = 0 (kept), 1 (repaired), 2 (failed)
-- @args feeders TEXT[]
DROP FUNCTION IF EXISTS ensure_feeders(); -- to allow changes to function signature
CREATE OR REPLACE FUNCTION ensure_feeders(hostkeys TEXT[]) RETURNS TEXT AS $ensure_hubs$
DECLARE
  hostkey TEXT;
  exit_code INT = 0;

BEGIN
  RAISE DEBUG 'hostkeys are %', hostkeys;
  FOREACH hostkey IN ARRAY hostkeys
  LOOP
    IF hostkey = '' THEN
      RAISE EXCEPTION 'hostkey cannot be empty';
    END IF;
    RAISE DEBUG 'make sure hostkey % is present in __hubs', hostkey;
    BEGIN
      INSERT INTO __hubs (hostkey) VALUES (hostkey);
      exit_code = 1;
    EXCEPTION
    WHEN unique_violation THEN
      RAISE DEBUG '% is already present in __hubs', hostkey;
    WHEN others THEN
      exit_code = 2;
    END;
  END LOOP;
  RETURN 'exit_code=' || exit_code;
END;
$ensure_hubs$ LANGUAGE plpgsql;

-- ENT-4862 Create view to select promises not kept (failed) that have not been kept or repaired
CREATE OR REPLACE FUNCTION ensure_not_kept_not_repaired_view() RETURNS VOID AS $body$
BEGIN
  EXECUTE 'CREATE OR REPLACE VIEW not_kept_not_repaired WITH (security_barrier) AS
    WITH p AS (
      SELECT * FROM __promiselog
      WHERE  promiseoutcome = ''NOTKEPT''
      AND NOT EXISTS (
        SELECT 1 FROM __promiseexecutions
        WHERE __promiselog.promisehash = __promiseexecutions.promisehash
        )
      )
      SELECT p.* FROM p'
      || get_rbac_sql();
END;
$body$ LANGUAGE plpgsql;

SELECT ensure_view('FUNCTION', 'ensure_not_kept_not_repaired_view');

COMMENT ON VIEW not_kept_not_repaired IS 'View to select promises not kept (failed) that have not been kept or repaired';
